from ..constraint import Constraint
from ..set import Set
from ..constants import EPSILON


class PerspectiveCone(Set):
    def __init__(self, func, expr, p):
        super().__init__("PerspectiveCone")
        self.func = func
        self.expr = expr
        self.p = p

    def constrain(self, var):
        return Constraint(var, "==", self.p * self.func(self.expr / (self.p + EPSILON)))



import threading
import logging

import autograd.numpy as np

from .expression import BaseExpr
from .constraint import Constraint
from .constants import Curvature as C

# Thread-local storage for variable ID counter
_thread_local = threading.local()
logger = logging.getLogger(__name__)


def _get_next_id() -> int:
    """Get the next variable ID in a thread-safe manner."""
    if not hasattr(_thread_local, "var_id"):
        _thread_local.var_id = 0
    current = _thread_local.var_id
    _thread_local.var_id += 1
    return current


def reset_variable_ids() -> None:
    """Reset the variable ID counter for the current thread."""
    _thread_local.var_id = 0


class Variable(BaseExpr):
    __array_priority__ = 100

    def __init__(
        self,
        shape: tuple[int, ...] = (1,),
        name: str | None = None,
        symmetric: bool = False,
        PSD: bool = False,
        NSD: bool = False,
        pos: bool = False,
        neg: bool = False,
        binary: bool = False,
        integer: bool = False,
    ) -> None:
        if not isinstance(shape, tuple):
            raise TypeError("Shape must be a tuple")
        if len(shape) == 0:
            raise ValueError("Shape must be non-empty")
        if not all(isinstance(s, int) for s in shape):
            raise TypeError("Shape must be a tuple of integers")
        if not all(s > 0 for s in shape):
            raise ValueError("Shape must be a tuple of positive integers")
        if len(shape) not in (1, 2):
            raise ValueError("Shape must be a tuple of length 1 or 2")

        var_id = _get_next_id()
        self.name = name if name else f"x{var_id}"
        self.shape = shape
        self.size = int(np.prod(shape))
        self._value = None
        self._id = var_id

        self.constraints = []

        if symmetric:
            if len(shape) != 2 or shape[0] != shape[1]:
                raise ValueError("Symmetric variable must be a square matrix")
            U_inds = np.triu_indices(shape[0], k=1)
            L_inds = np.tril_indices(shape[0], k=-1)
            self.constraints.append(Constraint(self[U_inds], "==", self[L_inds]))

        if PSD:
            if len(shape) != 2 or shape[0] != shape[1]:
                raise ValueError("PSD variable must be a square matrix")
            if neg:
                raise ValueError("PSD variable cannot be negative")
            self.constraints.append(Constraint(self, ">>", 0))

        if NSD:
            if len(shape) != 2 or shape[0] != shape[1]:
                raise ValueError("NSD variable must be a square matrix")
            if pos:
                raise ValueError("NSD variable cannot be positive")
            self.constraints.append(Constraint(self, "<<", 0))

        if pos:
            if neg or NSD:
                raise ValueError("Positive variable cannot be NSD or negative")
            self.constraints.append(Constraint(self, ">=", 0))

        if neg:
            if PSD or pos:
                raise ValueError("Negative variable cannot be PSD or positive")
            self.constraints.append(Constraint(self, "<=", 0))

        if binary:
            if neg or PSD or NSD:
                raise ValueError("Binary variable cannot be negative, PSD, or NSD")
            if pos:
                logger.warning("Setting binary variable to be positive is redundant, can only be 0 or 1")
            if integer:
                logger.warning("Setting binary variable to be integer is redundant")
            self.constraints.append(Constraint(self, ">=", 0))
            self.constraints.append(Constraint(self, "<=", 1))

        self.is_integer = bool(integer) or bool(binary)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        arr = np.array(val)
        if arr.size != self.size:
            raise ValueError(f"Cannot assign value with {arr.size} elements to variable with shape {self.shape} ({self.size} elements)")
        self._value = arr.reshape(self.shape)

    def __repr__(self) -> str:
        return f"Var({self.name}, shape={self.shape})"

    def __hash__(self) -> int:
        return hash(str(self))
    
    @property
    def curvature(self):
        return C.AFFINE



from __future__ import annotations

import autograd.numpy as np

from ..expression import Expr, BaseExpr, ExprLike
from ..constants import Curvature as C


class log(Expr):
    """Natural logarithm.

    log(x) is concave on x > 0.

    Curvature rules:
    - log(constant) -> constant
    - log(affine) -> concave
    - log(concave) -> concave (composition rule: concave + nondecreasing)
    - log(convex) -> unknown
    """

    def __init__(self, left: ExprLike) -> None:
        super().__init__("log", left)

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return np.log(x)

    @property
    def curvature(self) -> C:
        arg = self.left
        arg_cvx = arg.curvature if isinstance(arg, BaseExpr) else C.CONSTANT

        if arg_cvx == C.CONSTANT:
            return C.CONSTANT

        if arg_cvx in (C.AFFINE, C.CONCAVE):
            # log is concave and nondecreasing, so log(concave) is concave
            return C.CONCAVE

        return C.UNKNOWN

    @property
    def shape(self) -> tuple[int, ...]:
        if isinstance(self.left, BaseExpr):
            return self.left.shape
        return np.shape(self.left) or (1,)



from __future__ import annotations

from typing import Any, Iterable

import autograd.numpy as np

from .expression import Expr, ExprLike
from .variable import Variable
from .constructs.function import Function


def collect_vars(expr: ExprLike, vars: list[Variable]) -> None:
    """Recursively collect all Variable instances from an expression tree."""
    if isinstance(expr, Variable):
        vars.append(expr)
    elif isinstance(expr, Function):
        for arg in expr.args:
            collect_vars(arg, vars)
    elif isinstance(expr, Expr):
        collect_vars(expr.left, vars)
        if expr.right is not None:
            collect_vars(expr.right, vars)
    elif isinstance(expr, dict):
        for v in expr.values():
            collect_vars(v, vars)
    elif isinstance(expr, Iterable) and not isinstance(expr, np.ndarray):
        for e in expr:
            collect_vars(e, vars)


def eval_expression(
    expr: ExprLike, var_dict: dict[str, np.ndarray] | None, use_value: bool = False
) -> Any:
    """Evaluate an expression tree by substituting variable values."""
    if isinstance(expr, Variable):
        if use_value:
            return expr.value
        else:
            return var_dict[expr.name]

    elif isinstance(expr, Function):
        args = []
        for arg in expr.args:
            args.append(eval_expression(arg, var_dict, use_value))
        return expr.func(*args, **expr.kwargs)

    elif isinstance(expr, Expr):
        left_eval = eval_expression(expr.left, var_dict, use_value)
        right_eval = (
            eval_expression(expr.right, var_dict, use_value)
            if expr.right is not None
            else None
        )

        if expr.op == "add":
            return left_eval + right_eval
        elif expr.op == "sub":
            return left_eval - right_eval
        elif expr.op == "mul":
            return left_eval * right_eval
        elif expr.op == "div":
            return left_eval / right_eval
        elif expr.op == "pow":
            return left_eval**right_eval
        elif expr.op == "neg":
            return -left_eval
        elif expr.op == "matmul":
            return left_eval @ right_eval
        elif expr.op == "getitem":
            return left_eval[right_eval]
        elif expr.op == "transpose":
            return left_eval.T
        elif expr.op == "flatten":
            return left_eval.flatten()
        elif callable(expr):
            # Handle callable Expr subclasses (atoms like norm, abs, etc.)
            return expr(left_eval) if right_eval is None else expr(left_eval, right_eval)
        else:
            raise NotImplementedError(expr.op)

    elif isinstance(expr, dict):
        return {k: eval_expression(v, var_dict, use_value) for k, v in expr.items()}

    elif isinstance(expr, Iterable) and not isinstance(expr, np.ndarray):
        container = type(expr)
        return container([eval_expression(e, var_dict, use_value) for e in expr])

    else:
        return expr



from __future__ import annotations

import autograd.numpy as np

from ..expression import Expr, BaseExpr, ExprLike
from ..constants import Curvature as C


class amin(Expr):
    """Minimum value along an axis.

    amin(x) returns the minimum element of x (scalar if axis=None).

    Curvature rules:
    - amin(constant) -> constant
    - amin(affine) -> concave
    - amin(concave) -> concave
    - amin(convex) -> unknown
    """

    def __init__(self, left: ExprLike, axis: int | None = None) -> None:
        self.axis = axis
        super().__init__("amin", left)

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return np.amin(x, axis=self.axis)

    @property
    def curvature(self) -> C:
        arg = self.left
        arg_cvx = arg.curvature if isinstance(arg, BaseExpr) else C.CONSTANT

        if arg_cvx == C.CONSTANT:
            return C.CONSTANT

        if arg_cvx in (C.AFFINE, C.CONCAVE):
            return C.CONCAVE

        return C.UNKNOWN

    @property
    def shape(self) -> tuple[int, ...]:
        if self.axis is None:
            return (1,)
        arg_shape = self.left.shape if isinstance(self.left, BaseExpr) else np.shape(self.left)
        ndim = len(arg_shape)
        axis = self.axis if self.axis >= 0 else ndim + self.axis
        if axis < 0 or axis >= ndim:
            raise ValueError(f"axis {self.axis} is out of bounds for array of dimension {ndim}")
        result = list(arg_shape)
        result.pop(axis)
        return tuple(result) if result else (1,)


"""
Branch-and-Bound MINLP Backend

Implements a branch-and-bound algorithm for mixed-integer nonlinear programming
(MINLP) with multiple advanced features for robust solving.

Features:
- Multiple node selection strategies (best-first, depth-first, hybrid)
- Multiple branching strategies (most fractional, pseudocost, strong branching, reliability)
- Outer approximation (OA) cuts for convex constraint tightening (disabled by default)
- Primal heuristics (rounding, feasibility pump using scipy's MILP solver)
- Warm starting from parent solutions
- Node pruning by bound
- Configurable tolerances and limits
- Discrete value constraints (x ^ [1, 10, 100, ...])

Algorithm Details:
- Each node solves an NLP relaxation using scipy.optimize.minimize (default: SLSQP)
- The global lower bound (best_bound) is the minimum NLP relaxation objective
  across all open nodes in the queue
- Pseudocosts are learned from strong branching NLP solves when using
  reliability or strong branching strategies

Outer Approximation (OA) Cuts:
- For CONVEX problems only (set bb_assume_convex=True)
- Only inequality constraints are linearized; equality constraints are skipped.
- Constraint cuts for g(x) >= 0:
  For convex g, linearization gives: g(x*) + ∇g(x*)ᵀ(x - x*) >= 0
  => ∇g(x*)ᵀ x >= ∇g(x*)ᵀ x* - g(x*)
  These cuts are added as linear inequality constraints in node NLPs,
  tightening the feasible region.

WARNING: For non-convex problems, OA cuts can cut off the global optimum!
Only enable OA with bb_assume_convex=True if you are certain the problem is convex.

Note on solver character:
This is a "local" MINLP B&B because each node solves a local NLP (SLSQP) with
no guarantee of finding the global continuous relaxation optimum. For non-convex
problems, the bounds computed are valid lower bounds on local basins, making
this a heuristic global solver. For convex problems with OA constraint cuts,
the method provides valid tightening of the feasible region.
"""

from __future__ import annotations

import heapq
import logging
import time
from typing import Callable, Dict, List, Tuple

import autograd.numpy as np
from autograd import grad
from scipy.optimize import minimize

from ...constants import DEFAULT_INT_TOL

logger = logging.getLogger(__name__)

from ..base import build_objective, ProblemData, SolverResult, SolverStats, SolverStatus
from ..scipy_backend import ScipyBackend

from .node import (
    BBNode,
    BBStats,
    BranchingStrategy,
    NodeSelection,
    PseudocostData,
)
from .cuts import OACut, generate_oa_cuts, prune_cut_pool
from .branching import select_branching_variable
from .heuristics import rounding_heuristic, run_initial_heuristics
from .utils import (
    build_constraints_filtered,
    create_discrete_child_nodes,
    extract_discrete_constraints,
    extract_simple_bounds,
    get_integer_indices,
    get_integer_violations,
    get_scipy_bounds,
    get_warm_start,
    propagate_discrete_bounds,
    remove_redundant_equality_constraints,
)


class BranchAndBoundBackend:
    """
    Comprehensive branch-and-bound solver for mixed-integer nonlinear programs.

    Supports multiple search strategies, branching rules, and enhancement
    techniques for robust MINLP solving.
    """

    def solve(
        self,
        problem_data: ProblemData,
        solver: str,  # noqa: ARG002 - ignored, uses internal solvers
        solver_options: Dict[str, object],
    ) -> SolverResult:
        """
        Solve a MINLP using branch-and-bound.

        Args:
            problem_data: The problem specification
            solver: Ignored (uses internal NLP solver)
            solver_options: Options including B&B and NLP options:

                B&B options:
                - bb_max_nodes: Maximum nodes to explore (default: 10000)
                - bb_max_time: Maximum time in seconds (default: 300)
                - bb_abs_gap: Absolute optimality gap tolerance (default: 1e-6)
                - bb_rel_gap: Relative optimality gap tolerance (default: 1e-4)
                - bb_int_tol: Integer feasibility tolerance (default: 1e-5)
                - bb_verbose: Print progress (default: False)
                - bb_node_selection: "best_first", "depth_first", "hybrid" (default: "best_first")
                - bb_branching: "most_fractional", "pseudocost", "strong", "reliability" (default: "most_fractional")
                - bb_assume_convex: Assume problem is convex (enables OA cuts) (default: False)
                - bb_use_oa_cuts: Enable outer approximation cuts (default: bb_assume_convex)
                  WARNING: OA cuts can cut off optimal solutions for non-convex problems!
                - bb_use_heuristics: Enable primal heuristics (default: True)
                - bb_strong_branch_limit: Max strong branching candidates (default: 5)
                - bb_reliability_limit: Branches before pseudocost reliable (default: 8)
                - bb_max_cuts: Maximum OA cuts to keep in pool (default: 200)
                - bb_cut_max_age: Remove cuts older than this many nodes (default: 50)

                NLP solver options:
                - nlp_method: scipy.optimize.minimize method (default: "SLSQP")
                - nlp_maxiter: Maximum NLP iterations per solve (default: 1000)
                - nlp_ftol: NLP function tolerance (default: 1e-9)

                Feasibility pump options:
                - fp_max_iterations: Maximum FP iterations (default: 0, disabled)
                - fp_penalty_init: Initial penalty weight for distance term (default: 0.1)
                - fp_penalty_growth: Penalty multiplier each iteration (default: 1.5)
                - fp_use_oa: Use OA cuts in FP MILP subproblem (default: False)
                - fp_time_limit: Time limit for MILP subproblems in seconds (default: 0.5)

        Returns:
            SolverResult with the best solution found
        """
        start_time = time.time()
        setup_time = problem_data.setup_time

        # Extract B&B options
        options = dict(solver_options)
        max_nodes = int(options.pop("bb_max_nodes", 10000))
        max_time = float(options.pop("bb_max_time", 300))
        abs_gap = float(options.pop("bb_abs_gap", 1e-6))
        rel_gap = float(options.pop("bb_rel_gap", 1e-4))
        int_tol = float(options.pop("bb_int_tol", DEFAULT_INT_TOL))
        verbose = bool(options.pop("bb_verbose", False))

        # Configure logging for verbose mode
        if verbose:
            logger.setLevel(logging.INFO)
            if not logger.handlers and not logger.parent.handlers:
                handler = logging.StreamHandler()
                handler.setFormatter(logging.Formatter('%(message)s'))
                logger.addHandler(handler)

        # Strategy options (speed-first defaults)
        node_sel_opt = options.pop("bb_node_selection", "best_first")
        branch_opt = options.pop("bb_branching", "most_fractional")
        assume_convex = bool(options.pop("bb_assume_convex", False))
        use_oa_cuts = bool(options.pop("bb_use_oa_cuts", assume_convex))
        use_heuristics = bool(options.pop("bb_use_heuristics", True))
        strong_limit = int(options.pop("bb_strong_branch_limit", 5))
        reliability_limit = int(options.pop("bb_reliability_limit", 8))
        max_cuts = int(options.pop("bb_max_cuts", 200))
        cut_max_age = int(options.pop("bb_cut_max_age", 50))

        # NLP solver options
        nlp_maxiter = int(options.pop("nlp_maxiter", 1000))
        nlp_ftol = float(options.pop("nlp_ftol", 1e-9))
        nlp_method = str(options.pop("nlp_method", "SLSQP"))

        # Feasibility pump options
        fp_max_iterations = int(options.pop("fp_max_iterations", 0))
        fp_penalty_init = float(options.pop("fp_penalty_init", 0.1))
        fp_penalty_growth = float(options.pop("fp_penalty_growth", 1.5))
        fp_use_oa = bool(options.pop("fp_use_oa", False))
        fp_time_limit = float(options.pop("fp_time_limit", 0.5))

        # Remaining options are passed directly to the NLP solver
        nlp_options = dict(options)

        # Parse strategy enums
        node_selection = self._parse_enum(node_sel_opt, NodeSelection, "bb_node_selection")
        branching = self._parse_enum(branch_opt, BranchingStrategy, "bb_branching")

        # Extract discrete constraints
        discrete_vars, remaining_constraints = extract_discrete_constraints(problem_data)

        if verbose:
            logger.info(f"Problem size: {len(problem_data.x0)} variables")
            logger.info(f"Integer variables: {problem_data.integer_vars}")
            logger.info(f"Discrete constraints: {len(discrete_vars)}")
            logger.info(f"Total constraints: {len(problem_data.constraints)}")
            logger.info(f"Remaining constraints (after filtering discrete): {len(remaining_constraints)}")

        # If no integer variables and no discrete constraints, solve as NLP
        if not problem_data.integer_vars and not discrete_vars:
            if verbose:
                logger.info("No integer/discrete variables - solving as NLP directly")
            return self._solve_nlp(problem_data, nlp_options, nlp_method)

        # Initialize B&B statistics
        stats = BBStats()

        # Get integer variable indices
        int_indices = get_integer_indices(problem_data)

        # Add discrete variable indices (only non-pure-range)
        for idx, dvar in discrete_vars.items():
            if idx not in int_indices and not dvar.is_pure_range:
                int_indices.append(idx)
        int_indices = sorted(set(int_indices))

        int_indices_set = set(int_indices)
        n_vars = len(problem_data.x0)

        # Build objective and constraint functions
        obj_func = build_objective(problem_data)
        obj_grad = grad(obj_func)
        cons = build_constraints_filtered(problem_data, remaining_constraints)

        # Remove redundant equality constraints
        cons, n_removed = remove_redundant_equality_constraints(
            cons, problem_data.x0, verbose
        )

        # Initialize pseudocost data
        pseudocosts: Dict[int, PseudocostData] = {
            idx: PseudocostData() for idx in int_indices
        }

        # OA cuts pool
        oa_cuts: List[OACut] = []

        # Extract simple variable bounds
        simple_bounds = extract_simple_bounds(problem_data)

        # Initialize variable bounds
        initial_var_bounds = propagate_discrete_bounds(discrete_vars, simple_bounds)

        # Initialize best solution (incumbent)
        incumbent_x: np.ndarray | None = None
        incumbent_obj = float("inf")

        # Try to find initial solution via heuristics
        if use_heuristics:
            if verbose:
                logger.info("Running initial heuristics...")
            heur_x, heur_obj = run_initial_heuristics(
                problem_data, obj_func, cons, int_indices,
                nlp_method, nlp_maxiter, nlp_ftol, discrete_vars,
                fp_max_iterations, fp_penalty_init, fp_penalty_growth,
                fp_use_oa, fp_time_limit,
            )
            if heur_x is not None:
                incumbent_x = heur_x
                incumbent_obj = float(heur_obj)
                stats.heuristic_solutions += 1
                if verbose:
                    logger.info(f"Heuristic found initial solution: {incumbent_obj:.6e}")
            elif verbose:
                logger.info("Initial heuristics found no feasible solution")

        # Create root node
        root = BBNode(
            priority=float("-inf"),
            node_id=0,
            depth=0,
            var_bounds=initial_var_bounds,
            parent_solution=problem_data.x0.copy(),
            lower_bound=float("-inf"),
        )

        # Priority queue
        node_queue: List[BBNode] = [root]
        heapq.heapify(node_queue)

        node_counter = 1
        depth_first_counter = 0

        if verbose:
            n_discrete = len(discrete_vars)
            logger.info(f"Branch-and-Bound: {len(int_indices)} integer/discrete variable elements")
            if n_discrete > 0:
                logger.info(f"  ({n_discrete} with discrete value constraints)")
            logger.info(f"Strategy: {node_selection.value}, Branching: {branching.value}")
            logger.info(f"OA cuts: {use_oa_cuts}, Heuristics: {use_heuristics}")
            logger.info(f"Max nodes: {max_nodes}, Max time: {max_time}s")
            logger.info(f"{'Nodes':>8} {'Incumbent':>12} {'Best Bound':>12} {'Gap':>10} {'Time':>8}")
            logger.info("-" * 54)

        # Main B&B loop
        while node_queue:
            # Check termination conditions
            elapsed = time.time() - start_time
            if elapsed > max_time:
                if verbose:
                    logger.info(f"Time limit reached ({max_time}s)")
                break

            if stats.nodes_explored >= max_nodes:
                if verbose:
                    logger.info(f"Node limit reached ({max_nodes})")
                break

            # Update best bound from queue
            if node_queue:
                finite_bounds = [n.lower_bound for n in node_queue if np.isfinite(n.lower_bound)]
                if finite_bounds:
                    stats.best_bound = min(finite_bounds)

            # Check gap
            if incumbent_x is not None and np.isfinite(stats.best_bound) and np.isfinite(incumbent_obj):
                if abs(incumbent_obj) > 1e-10:
                    stats.gap = abs(incumbent_obj - stats.best_bound) / abs(incumbent_obj)
                else:
                    stats.gap = abs(incumbent_obj - stats.best_bound)

                if stats.gap <= rel_gap or abs(incumbent_obj - stats.best_bound) <= abs_gap:
                    if verbose:
                        logger.info(f"Optimality gap reached (gap={stats.gap:.2e})")
                    break

            # Select node
            node = self._select_node(node_queue, node_selection, depth_first_counter)
            depth_first_counter += 1
            stats.nodes_explored += 1

            # Prune by bound
            if node.lower_bound >= incumbent_obj - abs_gap:
                stats.nodes_pruned += 1
                continue

            # Solve NLP relaxation
            x0 = get_warm_start(node, problem_data)
            bounds = get_scipy_bounds(node, n_vars)

            nlp_result = self._solve_node_nlp(
                obj_func, obj_grad, x0, bounds, cons, oa_cuts,
                nlp_method, nlp_maxiter, nlp_ftol
            )
            stats.nlp_solves += 1

            if nlp_result is None:
                stats.nodes_infeasible += 1
                continue

            x_relaxed, obj_relaxed = nlp_result

            # Update node's bound
            node.lower_bound = obj_relaxed
            node.priority = obj_relaxed

            # Prune by bound
            if obj_relaxed >= incumbent_obj - abs_gap:
                stats.nodes_pruned += 1
                continue

            # Add OA cuts
            if use_oa_cuts:
                new_cuts = generate_oa_cuts(x_relaxed, cons)
                oa_cuts.extend(new_cuts)
                stats.cuts_added += len(new_cuts)

                for cut in oa_cuts:
                    cut.age += 1
                oa_cuts = prune_cut_pool(oa_cuts, max_cuts, cut_max_age)

            # Check integer feasibility
            int_violations = get_integer_violations(
                x_relaxed, int_indices, int_tol, discrete_vars, node.var_bounds
            )

            if not int_violations:
                # Integer feasible
                if obj_relaxed < incumbent_obj:
                    incumbent_x = x_relaxed.copy()
                    incumbent_obj = obj_relaxed

                    # Prune nodes with worse bounds
                    node_queue = [n for n in node_queue
                                  if n.lower_bound < incumbent_obj - abs_gap]
                    heapq.heapify(node_queue)
            else:
                # Try heuristics
                if use_heuristics and stats.nodes_explored % 10 == 0:
                    heur_x, heur_obj = rounding_heuristic(
                        x_relaxed, int_indices, int_indices_set,
                        problem_data, obj_func, cons,
                        nlp_method, nlp_maxiter, nlp_ftol,
                        discrete_vars
                    )
                    if heur_x is not None and float(heur_obj) < incumbent_obj:
                        incumbent_x = heur_x
                        incumbent_obj = float(heur_obj)
                        stats.heuristic_solutions += 1

                # Select branching variable
                sb_maxiter = max(50, nlp_maxiter // 10)
                sb_ftol = nlp_ftol * 10
                branch_idx, branch_val = select_branching_variable(
                    int_violations, pseudocosts, branching,
                    x_relaxed, obj_relaxed, obj_func, obj_grad, bounds, cons,
                    strong_limit, reliability_limit, stats,
                    nlp_method, sb_maxiter, sb_ftol, discrete_vars
                )

                # Create child nodes
                left_node, right_node = create_discrete_child_nodes(
                    node, branch_idx, branch_val, obj_relaxed,
                    x_relaxed, node_counter, discrete_vars
                )
                node_counter += 2

                heapq.heappush(node_queue, left_node)
                heapq.heappush(node_queue, right_node)

            # Progress output
            if verbose:
                elapsed_now = time.time() - start_time
                bound_str = f"{stats.best_bound:>12.4e}" if stats.best_bound > -1e30 else "        -inf"
                inc_str = f"{incumbent_obj:>12.4e}" if incumbent_x is not None else "         inf"
                logger.info(f"{stats.nodes_explored:>8} {inc_str} "
                      f"{bound_str} {stats.gap:>10.2e} "
                      f"{elapsed_now:>7.1f}s")

        # Determine final status
        solve_time = time.time() - start_time

        if not node_queue and incumbent_x is not None:
            stats.best_bound = incumbent_obj
            stats.gap = 0.0

        if incumbent_x is None:
            status = SolverStatus.INFEASIBLE
            x_sol = problem_data.x0
        elif stats.gap <= rel_gap or (stats.best_bound > -1e30 and
                                       abs(incumbent_obj - stats.best_bound) <= abs_gap):
            status = SolverStatus.OPTIMAL
            x_sol = incumbent_x
        else:
            status = SolverStatus.SUBOPTIMAL
            x_sol = incumbent_x

        if verbose:
            logger.info("-" * 54)
            logger.info(f"Status: {status}")
            logger.info(f"Nodes explored: {stats.nodes_explored}")
            logger.info(f"NLP solves: {stats.nlp_solves}")
            if stats.cuts_added > 0:
                logger.info(f"OA cuts added: {stats.cuts_added}")
            logger.info(f"Heuristic solutions: {stats.heuristic_solutions}")
            if incumbent_x is not None:
                logger.info(f"Best objective: {incumbent_obj:.6e}")
                if stats.best_bound > -1e30:
                    logger.info(f"Best bound: {stats.best_bound:.6e}")
                    logger.info(f"Gap: {stats.gap:.2e}")

        solver_stats = SolverStats(
            solver_name="B&B(SLSQP)",
            solve_time=solve_time,
            setup_time=setup_time,
            num_iters=stats.nodes_explored,
        )

        return SolverResult(
            x=x_sol,
            status=status,
            stats=solver_stats,
            raw_result={
                "bb_stats": stats,
                "incumbent_obj": incumbent_obj if incumbent_x is not None else None,
                "best_bound": stats.best_bound,
                "gap": stats.gap,
            },
        )

    def _select_node(
        self,
        node_queue: List[BBNode],
        strategy: NodeSelection,
        counter: int,
    ) -> BBNode:
        """Select next node to process based on strategy."""
        if strategy == NodeSelection.BEST_FIRST:
            return heapq.heappop(node_queue)

        elif strategy == NodeSelection.DEPTH_FIRST:
            max_depth = -1
            max_idx = 0
            for i, node in enumerate(node_queue):
                if node.depth > max_depth:
                    max_depth = node.depth
                    max_idx = i
            node = node_queue.pop(max_idx)
            heapq.heapify(node_queue)
            return node

        else:  # HYBRID
            if counter % 10 == 0:
                return heapq.heappop(node_queue)
            else:
                max_depth = -1
                max_idx = 0
                for i, node in enumerate(node_queue):
                    if node.depth > max_depth:
                        max_depth = node.depth
                        max_idx = i
                node = node_queue.pop(max_idx)
                heapq.heapify(node_queue)
                return node

    def _solve_nlp(
        self,
        problem_data: ProblemData,
        options: Dict,
        method: str = "SLSQP",
    ) -> SolverResult:
        """Solve a pure NLP (no integer variables)."""
        backend = ScipyBackend()
        return backend.solve(problem_data, method, options)

    def _solve_node_nlp(
        self,
        obj_func: Callable,
        obj_grad: Callable,
        x0: np.ndarray,
        bounds: List[Tuple[float | None, float | None]],
        cons: List[Dict],
        oa_cuts: List[OACut],
        method: str = "SLSQP",
        maxiter: int = 1000,
        ftol: float = 1e-9,
    ) -> Tuple[np.ndarray, float] | None:
        """Solve NLP relaxation at a node."""
        # Add OA cuts as linear constraints
        all_cons = list(cons)

        for cut in oa_cuts:
            if not cut.is_equality:
                def make_cut_fun(c):
                    def cut_fun(x):
                        return np.dot(c.coefficients, x) - c.rhs
                    return cut_fun

                all_cons.append({
                    "type": "ineq",
                    "fun": make_cut_fun(cut),
                })

        try:
            minimize_kwargs = {
                "method": method,
                "bounds": bounds,
                "constraints": all_cons,
                "options": {"maxiter": maxiter, "ftol": ftol},
            }
            if method in ScipyBackend.GRADIENT_METHODS:
                minimize_kwargs["jac"] = obj_grad
            result = minimize(obj_func, x0, **minimize_kwargs)

            if not result.success:
                return None

            x_sol = result.x

            # Verify constraint feasibility
            con_tol = 1e-4
            feasible = True
            max_violation = 0.0
            for con in cons:
                try:
                    con_val = con["fun"](x_sol)
                    if con["type"] == "eq":
                        violation = np.max(np.abs(con_val))
                        if violation > con_tol:
                            feasible = False
                            max_violation = max(max_violation, violation)
                    else:
                        violation = -np.min(con_val)
                        if violation > con_tol:
                            feasible = False
                            max_violation = max(max_violation, violation)
                except Exception as e:
                    logger.warning(f"Constraint evaluation failed: {e}")
                    feasible = False
                    break

            if not feasible:
                logger.debug(f"Node NLP rejected: constraint violation = {max_violation:.2e}")
                return None

            # Track active cuts
            for cut in oa_cuts:
                slack = np.dot(cut.coefficients, x_sol) - cut.rhs
                if abs(slack) < 1e-4:
                    cut.times_active += 1

            fun = result.fun
            if np.ndim(fun) > 0:
                fun = np.ravel(fun)[0]
            return x_sol, float(fun)

        except Exception as e:
            logger.warning(f"Node NLP solve failed: {e}")
            return None

    def _parse_enum(self, value, enum_class, option_name: str):
        """Parse an enum value from string or enum instance."""
        if isinstance(value, enum_class):
            return value
        if isinstance(value, str):
            try:
                return enum_class(value)
            except ValueError:
                pass
            value_upper = value.upper()
            for member in enum_class:
                if member.name == value_upper or member.value.upper() == value_upper:
                    return member
            valid = [m.value for m in enum_class]
            raise ValueError(
                f"Invalid value '{value}' for {option_name}. "
                f"Valid options: {valid}"
            )
        raise ValueError(
            f"Invalid type for {option_name}: expected string or {enum_class.__name__}, "
            f"got {type(value).__name__}"
        )




"""Direct IPOPT backend using cyipopt."""

from __future__ import annotations

import logging
import time
import warnings
from typing import Dict

import autograd.numpy as np
from autograd import grad, jacobian

from ..parser import eval_expression
from ..compiler import compile_to_function
from ..constants import EPSILON, DEFAULT_SOLVER_TOL

logger = logging.getLogger(__name__)
from .base import (
    build_objective,
    uses_projection,
    ProblemData,
    SolverResult,
    SolverStats,
    SolverStatus,
)


class IpoptBackend:
    """Backend for IPOPT solver using cyipopt directly."""

    def solve(
        self,
        problem_data: ProblemData,
        solver: str,
        solver_options: Dict[str, object],
    ) -> SolverResult:
        try:
            import cyipopt
        except ImportError as exc:
            raise ImportError(
                "IPOPT backend requires 'cyipopt' to be installed. "
                "Install with: pip install cyipopt"
            ) from exc

        if problem_data.integer_vars:
            raise ValueError(
                "IPOPT does not support integer variables. "
                "Use nvx.BNB for mixed-integer problems."
            )

        x0 = np.asarray(problem_data.x0, dtype=float)
        n = len(x0)
        setup_time = problem_data.setup_time

        # Build objective and gradient
        compile_start = time.time()
        obj_func = build_objective(problem_data)
        obj_grad = grad(obj_func)

        # Build constraints
        eq_constraints, ineq_constraints = self._build_constraints(
            problem_data, for_projection=False
        )
        setup_time += time.time() - compile_start

        # Count constraint dimensions
        n_eq = sum(self._constraint_dim(c, x0) for c in eq_constraints)
        n_ineq = sum(self._constraint_dim(c, x0) for c in ineq_constraints)
        m = n_eq + n_ineq

        # Build combined constraint function and jacobian
        if m > 0:
            con_func, con_jac = self._build_combined_constraints(
                eq_constraints, ineq_constraints, n, m
            )
        else:
            def con_func(x):
                return np.array([])

            def con_jac(x):
                return np.zeros((0, n))

        # Constraint bounds: equality constraints have cl=cu=0, inequality have cl=0, cu=inf
        cl = np.zeros(m)
        cu = np.concatenate([
            np.zeros(n_eq),  # equality: g(x) = 0
            np.full(n_ineq, np.inf),  # inequality: g(x) >= 0
        ]) if m > 0 else np.array([])

        # Variable bounds (default: unbounded)
        lb = solver_options.get("lb", np.full(n, -1e20))
        ub = solver_options.get("ub", np.full(n, 1e20))
        if isinstance(lb, (int, float)):
            lb = np.full(n, lb)
        if isinstance(ub, (int, float)):
            ub = np.full(n, ub)

        # Create the problem
        nlp = cyipopt.Problem(
            n=n,
            m=m,
            problem_obj=_IpoptProblem(obj_func, obj_grad, con_func, con_jac),
            lb=lb,
            ub=ub,
            cl=cl,
            cu=cu,
        )

        # Set IPOPT options
        nlp.add_option("print_level", solver_options.get("print_level", 0))
        nlp.add_option("max_iter", solver_options.get("maxiter", 1000))
        nlp.add_option("tol", solver_options.get("tol", DEFAULT_SOLVER_TOL))

        # Additional user options
        for key, value in solver_options.items():
            if key not in {"lb", "ub", "print_level", "maxiter", "tol"}:
                try:
                    nlp.add_option(key, value)
                except Exception as e:
                    logger.warning(f"Invalid option '{key}={value}': {e}")

        # Solve
        start_time = time.time()
        x_sol, info = nlp.solve(x0)
        solve_time = time.time() - start_time

        # Projection phase for <- constraints
        projection_info = None
        if uses_projection(problem_data):
            compile_start = time.time()
            proj_eq, proj_ineq = self._build_constraints(
                problem_data, for_projection=True
            )
            setup_time += time.time() - compile_start

            n_proj_eq = sum(self._constraint_dim(c, x_sol) for c in proj_eq)
            n_proj_ineq = sum(self._constraint_dim(c, x_sol) for c in proj_ineq)
            m_proj = n_proj_eq + n_proj_ineq

            if m_proj > 0:
                proj_con_func, proj_con_jac = self._build_combined_constraints(
                    proj_eq, proj_ineq, n, m_proj
                )

                proj_cl = np.zeros(m_proj)
                proj_cu = np.concatenate([
                    np.zeros(n_proj_eq),
                    np.full(n_proj_ineq, np.inf),
                ]) if m_proj > 0 else np.array([])

                # Dummy objective for projection (just find feasible point)
                def dummy_obj(_):
                    return 0.0

                def dummy_grad(x):
                    return np.zeros_like(x)

                proj_nlp = cyipopt.Problem(
                    n=n,
                    m=m_proj,
                    problem_obj=_IpoptProblem(dummy_obj, dummy_grad, proj_con_func, proj_con_jac),
                    lb=lb,
                    ub=ub,
                    cl=proj_cl,
                    cu=proj_cu,
                )

                proj_nlp.add_option("print_level", 0)
                proj_nlp.add_option("max_iter", problem_data.projection_maxiter)
                proj_nlp.add_option("tol", solver_options.get("tol", DEFAULT_SOLVER_TOL))

                start_time = time.time()
                proj_x_sol, projection_info = proj_nlp.solve(x_sol)
                solve_time += time.time() - start_time

                if projection_info.get("status", -100) in [0, 1, 6]:
                    x_sol = proj_x_sol
                else:
                    warnings.warn(
                        f"Projection step failed with status {projection_info.get('status')}"
                    )

        # Interpret status
        status = self._interpret_status(info, projection_info)

        stats = SolverStats(
            solver_name="IPOPT",
            solve_time=solve_time,
            setup_time=setup_time,
            num_iters=None,  # cyipopt doesn't expose iteration count easily
        )

        return SolverResult(
            x=x_sol,
            status=status,
            stats=stats,
            raw_result={"primary": info, "projection": projection_info},
        )

    @staticmethod
    def _build_constraints(problem_data: ProblemData, for_projection: bool = False):
        """Build separate lists of equality and inequality constraint functions.

        Handles all constraint types:
        - >= : Greater than or equal (inequality)
        - <= : Less than or equal (inequality)
        - == : Equality
        - >> : Positive semidefinite - eigenvalues must be >= 0
        - << : Negative semidefinite - eigenvalues must be <= 0
        - <- : Projection constraint (soft inequality during solve, equality during projection)
        """
        eq_constraints = []
        ineq_constraints = []
        use_compile = problem_data.compile
        p_tol = problem_data.projection_tolerance

        for constraint in problem_data.constraints:
            if constraint.op == "in":
                # Discrete constraints not handled by IPOPT
                continue

            def make_con_fun(c, compile_exprs=use_compile):
                if compile_exprs:
                    compiled_left = compile_to_function(c.left)
                    compiled_right = compile_to_function(c.right)

                    def con_fun(x):
                        var_dict = problem_data.unpack(x)
                        lval = compiled_left(var_dict)
                        rval = compiled_right(var_dict)
                        if c.op == "<-":
                            if for_projection:
                                res = lval - rval
                            else:
                                # Use epsilon smoothing to avoid gradient singularity at zero
                                diff = np.atleast_2d(lval - rval)
                                res = p_tol - np.sqrt(np.sum(diff * diff) + EPSILON)
                        else:
                            res = (
                                lval - rval
                                if c.op in [">=", "==", ">>"]
                                else rval - lval
                            )
                        if c.op in [">>", "<<"]:
                            res = np.real(np.ravel(np.linalg.eig(res)[0]))
                        return np.ravel(res)
                else:
                    def con_fun(x):
                        var_dict = problem_data.unpack(x)
                        lval = eval_expression(c.left, var_dict)
                        rval = eval_expression(c.right, var_dict)
                        if c.op == "<-":
                            if for_projection:
                                res = lval - rval
                            else:
                                # Use epsilon smoothing to avoid gradient singularity at zero
                                diff = np.atleast_2d(lval - rval)
                                res = p_tol - np.sqrt(np.sum(diff * diff) + EPSILON)
                        else:
                            res = (
                                lval - rval
                                if c.op in [">=", "==", ">>"]
                                else rval - lval
                            )
                        if c.op in [">>", "<<"]:
                            res = np.real(np.ravel(np.linalg.eig(res)[0]))
                        return np.ravel(res)

                return con_fun

            con_fun = make_con_fun(constraint)

            if constraint.op == "==" or (for_projection and constraint.op == "<-"):
                eq_constraints.append(con_fun)
            else:  # >=, <=, >>, <<, or <- (main solve)
                ineq_constraints.append(con_fun)

        return eq_constraints, ineq_constraints

    @staticmethod
    def _constraint_dim(con_func, x0):
        """Get the dimension of a constraint function."""
        try:
            return len(con_func(x0))
        except Exception:
            return 1

    @staticmethod
    def _build_combined_constraints(eq_constraints, ineq_constraints, n, m):
        """Build combined constraint function and jacobian."""
        all_constraints = eq_constraints + ineq_constraints

        def combined_con(x):
            results = []
            for con in all_constraints:
                results.append(np.atleast_1d(con(x)))
            return np.concatenate(results) if results else np.array([])

        combined_jac = jacobian(combined_con)

        return combined_con, combined_jac

    @staticmethod
    def _interpret_status(info: dict, projection_info: dict = None) -> SolverStatus:
        """Interpret IPOPT return status.

        IPOPT ApplicationReturnStatus codes:
            Success (>= 0):
                0: Solve_Succeeded
                1: Solved_To_Acceptable_Level
                2: Infeasible_Problem_Detected
                3: Search_Direction_Becomes_Too_Small
                4: Diverging_Iterates
                5: User_Requested_Stop
                6: Feasible_Point_Found

            Error (< 0):
                -1: Maximum_Iterations_Exceeded
                -2: Restoration_Failed
                -3: Error_In_Step_Computation
                -4: Maximum_CpuTime_Exceeded
                -5: Maximum_WallTime_Exceeded
                -10: Not_Enough_Degrees_Of_Freedom
                -11: Invalid_Problem_Definition
                -12: Invalid_Option
                -13: Invalid_Number_Detected
        """
        status = info.get("status", -100)

        # Map status code to solver status
        if status == 0:  # Solve_Succeeded
            base_status = SolverStatus.OPTIMAL
        elif status == 1:  # Solved_To_Acceptable_Level
            base_status = SolverStatus.SUBOPTIMAL
        elif status == 2:  # Infeasible_Problem_Detected
            return SolverStatus.INFEASIBLE
        elif status == 3:  # Search_Direction_Becomes_Too_Small
            return SolverStatus.NUMERICAL_ERROR
        elif status == 4:  # Diverging_Iterates
            return SolverStatus.UNBOUNDED
        elif status == 5:  # User_Requested_Stop
            return SolverStatus.ERROR
        elif status == 6:  # Feasible_Point_Found (for square problems)
            base_status = SolverStatus.OPTIMAL
        elif status == -1:  # Maximum_Iterations_Exceeded
            return SolverStatus.MAX_ITERATIONS
        elif status == -2:  # Restoration_Failed
            return SolverStatus.NUMERICAL_ERROR
        elif status == -3:  # Error_In_Step_Computation
            return SolverStatus.NUMERICAL_ERROR
        elif status == -4:  # Maximum_CpuTime_Exceeded
            return SolverStatus.MAX_ITERATIONS
        elif status == -5:  # Maximum_WallTime_Exceeded
            return SolverStatus.MAX_ITERATIONS
        elif status == -10:  # Not_Enough_Degrees_Of_Freedom
            return SolverStatus.ERROR
        elif status == -11:  # Invalid_Problem_Definition
            return SolverStatus.ERROR
        elif status == -12:  # Invalid_Option
            return SolverStatus.ERROR
        elif status == -13:  # Invalid_Number_Detected
            return SolverStatus.NUMERICAL_ERROR
        else:
            return SolverStatus.UNKNOWN

        # If projection was attempted and failed, downgrade to suboptimal
        if projection_info is not None:
            proj_status = projection_info.get("status", -100)
            if proj_status not in [0, 1, 6]:
                return SolverStatus.SUBOPTIMAL

        return base_status


class _IpoptProblem:
    """Wrapper class that provides the interface expected by cyipopt."""

    def __init__(self, objective, gradient, constraints, jacobian):
        self._objective = objective
        self._gradient = gradient
        self._constraints = constraints
        self._jacobian = jacobian

    def objective(self, x):
        return self._objective(x)

    def gradient(self, x):
        return self._gradient(x)

    def constraints(self, x):
        return self._constraints(x)

    def jacobian(self, x):
        jac = self._jacobian(x)
        return jac.flatten()  # cyipopt expects flattened jacobian



"""
Utility Functions for Branch-and-Bound

This module contains utility functions shared across the B&B implementation,
including bound extraction, constraint handling, and node management.
"""

from __future__ import annotations

import logging
from typing import Dict, List, Tuple

import autograd.numpy as np
from autograd import jacobian

from ...constants import Curvature
from ..base import ProblemData
from .node import BBNode, DiscreteVarInfo

logger = logging.getLogger(__name__)


def extract_simple_bounds(problem_data: ProblemData) -> Dict[int, Tuple[float, float]]:
    """
    Extract simple variable bounds from constraints.

    Looks for constraints of the form:
    - var >= constant  (lower bound)
    - var <= constant  (upper bound)
    - constant <= var  (lower bound)
    - constant >= var  (upper bound)

    Returns a dict mapping flat index -> (lb, ub).
    """
    from ...variable import Variable

    var_bounds: Dict[int, Tuple[float, float]] = {}
    n_vars = len(problem_data.x0)

    # Initialize with no bounds
    for i in range(n_vars):
        var_bounds[i] = (float("-inf"), float("inf"))

    for constraint in problem_data.constraints:
        if constraint.op not in (">=", "<=", "=="):
            continue

        # Check if this is a simple bound: var op constant or constant op var
        left = constraint.left
        right = constraint.right

        var = None
        const = None
        is_var_on_left = False

        if isinstance(left, Variable) and isinstance(right, (int, float)):
            var = left
            const = float(right)
            is_var_on_left = True
        elif isinstance(right, Variable) and isinstance(left, (int, float)):
            var = right
            const = float(left)
            is_var_on_left = False
        else:
            continue

        if var.name not in problem_data.var_slices:
            continue

        start, end = problem_data.var_slices[var.name]

        # Determine bound type based on operator and which side var is on
        for idx in range(start, end):
            current_lb, current_ub = var_bounds[idx]

            if constraint.op == ">=":
                if is_var_on_left:
                    # var >= const -> lower bound
                    current_lb = max(current_lb, const)
                else:
                    # const >= var -> upper bound
                    current_ub = min(current_ub, const)
            elif constraint.op == "<=":
                if is_var_on_left:
                    # var <= const -> upper bound
                    current_ub = min(current_ub, const)
                else:
                    # const <= var -> lower bound
                    current_lb = max(current_lb, const)
            elif constraint.op == "==":
                # var == const -> fixed
                current_lb = max(current_lb, const)
                current_ub = min(current_ub, const)

            var_bounds[idx] = (current_lb, current_ub)

    # Remove entries with no actual bounds (both inf)
    return {
        idx: bounds for idx, bounds in var_bounds.items()
        if bounds[0] > float("-inf") or bounds[1] < float("inf")
    }


def propagate_discrete_bounds(
    discrete_vars: Dict[int, DiscreteVarInfo],
    simple_bounds: Dict[int, Tuple[float, float]] | None = None,
) -> Dict[int, Tuple[float, float]]:
    """
    Initialize tight bounds from discrete constraints and simple variable bounds.

    For each discrete variable, sets bounds to [min_allowed, max_allowed].
    Also incorporates simple bounds extracted from constraints.
    """
    var_bounds: Dict[int, Tuple[float, float]] = {}

    # First, add simple bounds
    if simple_bounds:
        var_bounds.update(simple_bounds)

    # Then, tighten with discrete bounds (these are typically tighter)
    for idx, dvar in discrete_vars.items():
        # Collect all lower/upper bounds from discrete values and ranges
        lb_candidates = list(dvar.allowed_values)
        ub_candidates = list(dvar.allowed_values)

        # Add range bounds
        for r_lb, r_ub in dvar.allowed_ranges:
            lb_candidates.append(r_lb)
            ub_candidates.append(r_ub)

        if lb_candidates:  # Has at least one value or range
            lb = min(lb_candidates)
            ub = max(ub_candidates)

            # Intersect with existing bounds
            if idx in var_bounds:
                existing_lb, existing_ub = var_bounds[idx]
                lb = max(lb, existing_lb)
                ub = min(ub, existing_ub)

            var_bounds[idx] = (lb, ub)

    return var_bounds


def get_integer_indices(problem_data: ProblemData) -> List[int]:
    """Get flat indices for all integer variable elements."""
    indices = []
    for var_name in problem_data.integer_vars:
        start, end = problem_data.var_slices[var_name]
        indices.extend(range(start, end))
    return indices


def get_warm_start(node: BBNode, problem_data: ProblemData) -> np.ndarray:
    """Get warm start, projected to node bounds."""
    if node.parent_solution is not None:
        x0 = node.parent_solution.copy()
    else:
        x0 = problem_data.x0.copy()

    # Project to node bounds
    for idx, (lb, ub) in node.var_bounds.items():
        x0[idx] = np.clip(x0[idx], lb, ub)

    return x0


def get_scipy_bounds(
    node: BBNode,
    n_vars: int,
) -> List[Tuple[float | None, float | None]]:
    """Get bounds for scipy.optimize.minimize."""
    bounds: List[Tuple[float | None, float | None]] = [(None, None)] * n_vars

    # Apply node-specific bounds for integer variables
    for idx, (lb, ub) in node.var_bounds.items():
        bounds[idx] = (lb, ub)

    return bounds


def get_integer_violations(
    x: np.ndarray,
    int_indices: List[int],
    tol: float,
    discrete_vars: Dict[int, DiscreteVarInfo] | None = None,
    node_bounds: Dict[int, Tuple[float, float]] | None = None,
) -> List[Tuple[int, float]]:
    """Get list of (index, value) for variables violating integrality/discreteness."""
    violations = []

    # Check integer variables
    for idx in int_indices:
        val = x[idx]

        # Check if this index has a discrete constraint
        if discrete_vars and idx in discrete_vars:
            dvar = discrete_vars[idx]
            # Pure discrete values - use standard check
            if not dvar.is_feasible(val):
                violations.append((idx, val))
        else:
            # Standard integer check
            frac = abs(val - round(val))
            if frac > tol:
                violations.append((idx, val))

    # Check pure range variables (DiscreteRanges) - these are NOT in int_indices
    # but still need branching until committed to a single range
    if discrete_vars:
        for idx, dvar in discrete_vars.items():
            if idx in int_indices:
                continue  # Already checked above

            val = x[idx]

            # Get current bounds for this variable
            if node_bounds and idx in node_bounds:
                current_lb, current_ub = node_bounds[idx]
            else:
                current_lb, current_ub = -1e8, 1e8

            # Count how many ranges are still reachable within current bounds
            reachable_ranges = 0
            for r_lb, r_ub in dvar.allowed_ranges:
                # Range overlaps with current bounds?
                if r_lb <= current_ub + tol and r_ub >= current_lb - tol:
                    reachable_ranges += 1

            # If multiple ranges are reachable, we need to branch
            if reachable_ranges > 1:
                violations.append((idx, val))
            # If only one range but value not in it, still a violation
            elif not dvar.is_feasible(val):
                violations.append((idx, val))

    return violations


def extract_discrete_constraints(
    problem_data: ProblemData,
) -> Tuple[Dict[int, DiscreteVarInfo], List]:
    """
    Extract discrete (x ^ [values]) and range (x ^ [[lb, ub], ...]) constraints from problem data.

    Handles both DiscreteSet (discrete values only) and DiscreteRanges
    (continuous ranges for disjunctive constraints).

    Returns:
        Tuple of:
        - Dict mapping flat index -> DiscreteVarInfo
        - List of remaining constraints (non-discrete)
    """
    from ...sets.discrete_set import DiscreteSet, DiscreteRanges
    from ...variable import Variable

    discrete_vars: Dict[int, DiscreteVarInfo] = {}
    remaining_constraints = []

    for constraint in problem_data.constraints:
        if constraint.op == "in" and isinstance(constraint.right, (DiscreteSet, DiscreteRanges)):
            var = constraint.left
            the_set = constraint.right

            # Get the variable name
            if isinstance(var, Variable):
                var_name = var.name
                if var_name in problem_data.var_slices:
                    start, end = problem_data.var_slices[var_name]

                    # Extract values and ranges depending on set type
                    if isinstance(the_set, DiscreteRanges):
                        allowed_values = ()
                        allowed_ranges = tuple((r.lb, r.ub) for r in the_set.ranges)
                    else:
                        # DiscreteSet - only discrete values
                        allowed_values = the_set.values
                        allowed_ranges = ()

                    # For each element of the variable, add discrete info
                    for idx in range(start, end):
                        discrete_vars[idx] = DiscreteVarInfo(
                            var_name=var_name,
                            flat_index=idx,
                            allowed_values=allowed_values,
                            allowed_ranges=allowed_ranges,
                            tolerance=the_set.tolerance,
                        )
        else:
            remaining_constraints.append(constraint)

    return discrete_vars, remaining_constraints


def round_to_discrete(
    x: np.ndarray,
    int_indices: List[int],
    discrete_vars: Dict[int, DiscreteVarInfo] | None = None,
) -> np.ndarray:
    """Round integer variables to nearest allowed values (or project to nearest range)."""
    x_rounded = x.copy()

    # Round integer variables
    for idx in int_indices:
        if discrete_vars and idx in discrete_vars:
            dvar = discrete_vars[idx]
            # Use the nearest() method which handles both discrete values and ranges
            x_rounded[idx] = dvar.nearest(x[idx])
        else:
            # Standard rounding
            x_rounded[idx] = round(x[idx])

    # Project pure range variables (DiscreteRanges) to nearest valid range
    if discrete_vars:
        for idx, dvar in discrete_vars.items():
            if idx in int_indices:
                continue  # Already handled above
            # Project to nearest value in any range
            x_rounded[idx] = dvar.nearest(x[idx])

    return x_rounded


def create_discrete_child_nodes(
    parent: BBNode,
    branch_idx: int,
    branch_val: float,
    parent_obj: float,
    parent_x: np.ndarray,
    node_counter: int,
    discrete_vars: Dict[int, DiscreteVarInfo] | None = None,
) -> Tuple[BBNode, BBNode]:
    """Create child nodes, respecting discrete allowed values and ranges."""
    current_lb, current_ub = parent.var_bounds.get(branch_idx, (-1e8, 1e8))

    if discrete_vars and branch_idx in discrete_vars:
        dvar = discrete_vars[branch_idx]
        tol = dvar.tolerance

        # Collect all disjuncts (discrete values + ranges) that overlap with current bounds
        # Each disjunct is represented as (lb, ub) - for discrete values, lb == ub
        disjuncts_in_range = []

        # Discrete values
        for v in dvar.allowed_values:
            if current_lb - tol <= v <= current_ub + tol:
                disjuncts_in_range.append((v, v))

        # Ranges (may be truncated by current bounds)
        for r_lb, r_ub in dvar.allowed_ranges:
            # Check if range overlaps with current bounds
            effective_lb = max(r_lb, current_lb)
            effective_ub = min(r_ub, current_ub)
            if effective_lb <= effective_ub + tol:
                disjuncts_in_range.append((effective_lb, effective_ub))

        # Sort by lower bound
        disjuncts_in_range.sort(key=lambda d: d[0])

        if len(disjuncts_in_range) <= 1:
            # Only one disjunct left, should not be branching
            left_bounds = dict(parent.var_bounds)
            right_bounds = dict(parent.var_bounds)
            if disjuncts_in_range:
                d_lb, d_ub = disjuncts_in_range[0]
                left_bounds[branch_idx] = (d_lb, d_ub)
                right_bounds[branch_idx] = (d_lb, d_ub)
            else:
                # No disjuncts - should be infeasible
                left_bounds[branch_idx] = (1e8, -1e8)
                right_bounds[branch_idx] = (1e8, -1e8)
        else:
            # Split disjuncts into two sets based on branch_val
            # Disjuncts entirely below/at branch_val go left
            # Disjuncts entirely above branch_val go right
            # Ranges straddling branch_val are split

            left_disjuncts = []
            right_disjuncts = []

            for d_lb, d_ub in disjuncts_in_range:
                if d_ub <= branch_val:
                    # Entirely below/at branch_val
                    left_disjuncts.append((d_lb, d_ub))
                elif d_lb > branch_val:
                    # Entirely above branch_val
                    right_disjuncts.append((d_lb, d_ub))
                else:
                    # Range straddles branch_val - split it
                    if d_lb <= branch_val:
                        left_disjuncts.append((d_lb, branch_val))
                    if branch_val < d_ub:
                        right_disjuncts.append((branch_val, d_ub))

            # Ensure both sides have something (fallback to splitting in half)
            if not left_disjuncts:
                mid = len(disjuncts_in_range) // 2
                left_disjuncts = disjuncts_in_range[:mid]
                right_disjuncts = disjuncts_in_range[mid:]
            elif not right_disjuncts:
                mid = len(disjuncts_in_range) // 2
                left_disjuncts = disjuncts_in_range[:mid]
                right_disjuncts = disjuncts_in_range[mid:]

            # Compute bounds for each child
            # Left child: from min lower bound to max upper bound of left disjuncts
            # This ensures the NLP solver respects the actual disjunct boundaries
            left_bounds = dict(parent.var_bounds)
            if left_disjuncts:
                left_min = min(d[0] for d in left_disjuncts)
                left_max = max(d[1] for d in left_disjuncts)
                left_bounds[branch_idx] = (left_min, left_max)
            else:
                left_bounds[branch_idx] = (1e8, -1e8)  # Infeasible

            # Right child: from min lower bound to max upper bound of right disjuncts
            right_bounds = dict(parent.var_bounds)
            if right_disjuncts:
                right_min = min(d[0] for d in right_disjuncts)
                right_max = max(d[1] for d in right_disjuncts)
                right_bounds[branch_idx] = (right_min, right_max)
            else:
                right_bounds[branch_idx] = (1e8, -1e8)  # Infeasible
    else:
        # Standard integer branching
        left_bounds = dict(parent.var_bounds)
        left_bounds[branch_idx] = (current_lb, np.floor(branch_val))

        right_bounds = dict(parent.var_bounds)
        right_bounds[branch_idx] = (np.ceil(branch_val), current_ub)

    # Use the actual NLP relaxation objective (parent_obj) as the lower bound
    # This is the correct bound from the parent's NLP solve, not inherited from grandparent
    left_node = BBNode(
        priority=parent_obj,  # Priority for heap ordering (lower = better for best-first)
        node_id=node_counter,
        depth=parent.depth + 1,
        var_bounds=left_bounds,
        parent_solution=parent_x.copy(),
        lower_bound=parent_obj,  # Correct: use parent's NLP objective as child's lower bound
    )

    right_node = BBNode(
        priority=parent_obj,
        node_id=node_counter + 1,
        depth=parent.depth + 1,
        var_bounds=right_bounds,
        parent_solution=parent_x.copy(),
        lower_bound=parent_obj,
    )

    return left_node, right_node


def build_constraints_filtered(
    problem_data: ProblemData, 
    constraints: List
) -> List[Dict]:
    """Build scipy constraint dictionaries from a filtered list of constraints."""
    from ...parser import eval_expression
    from ...compiler import compile_to_function
    
    cons = []
    use_compile = problem_data.compile

    for constraint in constraints:
        # Skip "in" constraints (discrete set membership) - handled by B&B
        if constraint.op == "in":
            continue

        def make_con_fun(c, compile_exprs=use_compile):
            if compile_exprs:
                compiled_left = compile_to_function(c.left)
                compiled_right = compile_to_function(c.right)

                def con_fun(x):
                    var_dict = problem_data.unpack(x)
                    lval = compiled_left(var_dict)
                    rval = compiled_right(var_dict)
                    res = (
                        lval - rval
                        if c.op in [">=", "==", ">>"]
                        else rval - lval
                    )
                    return np.ravel(res)
            else:
                def con_fun(x):
                    var_dict = problem_data.unpack(x)
                    lval = eval_expression(c.left, var_dict)
                    rval = eval_expression(c.right, var_dict)
                    res = (
                        lval - rval
                        if c.op in [">=", "==", ">>"]
                        else rval - lval
                    )
                    return np.ravel(res)

            return con_fun

        con_fun = make_con_fun(constraint)
        con_jac = jacobian(con_fun)
        con_type = "eq" if constraint.op == "==" else "ineq"

        cons.append({
            "type": con_type,
            "fun": con_fun,
            "jac": con_jac,
            "curvature": constraint.curvature,
        })

    return cons


def remove_redundant_equality_constraints(
    cons: List[Dict],
    x0: np.ndarray,
    verbose: bool = False,
    tol: float = 1e-10,
) -> Tuple[List[Dict], int]:
    """
    Remove linearly dependent AFFINE equality constraints.

    Only processes constraints that are affine (constant Jacobian).
    Nonlinear equality constraints are never removed since their
    Jacobian varies with x, making point-wise redundancy checks invalid.

    Uses QR decomposition with column pivoting to identify and remove
    redundant affine constraints. This prevents singular Jacobian errors
    in scipy solvers like SLSQP.

    This is common in graph problems where degree constraints are
    naturally redundant (e.g., sum of in-degrees = sum of out-degrees).

    Args:
        cons: List of scipy constraint dictionaries (with 'curvature' key)
        x0: Initial point for Jacobian evaluation
        verbose: Whether to log information about removed constraints
        tol: Tolerance for determining linear dependence

    Returns:
        Tuple of (filtered constraints, number removed)
    """
    # Separate constraints by type and curvature
    # Only affine equality constraints are candidates for redundancy removal
    affine_eq_cons = []
    other_cons = []

    for i, c in enumerate(cons):
        curvature = c.get("curvature")
        is_affine = curvature in (Curvature.CONSTANT, Curvature.AFFINE)

        if c["type"] == "eq" and is_affine:
            affine_eq_cons.append((i, c))
        else:
            other_cons.append(c)

    if len(affine_eq_cons) <= 1:
        return cons, 0

    # Build Jacobian matrix for affine equality constraints only
    try:
        jac_rows = []
        eq_indices = []
        for orig_idx, c in affine_eq_cons:
            jac = c["jac"](x0)
            if jac.ndim == 1:
                jac = jac.reshape(1, -1)
            for row in jac:
                jac_rows.append(row)
                eq_indices.append(orig_idx)

        if not jac_rows:
            return cons, 0

        from scipy.linalg import qr as scipy_qr

        J = np.vstack(jac_rows)
        n_rows = J.shape[0]

        # Use QR decomposition with column pivoting on J^T to find row rank
        # pivoting=True returns permutation P where P[:rank] are independent rows
        _, R, P = scipy_qr(J.T, pivoting=True)

        # Find rank by counting significant diagonal elements of R
        diag_R = np.abs(np.diag(R))
        max_diag = np.max(diag_R) if len(diag_R) > 0 else 0
        if max_diag > 0:
            rank = int(np.sum(diag_R > tol * max_diag))
        else:
            rank = 0

        if rank == n_rows:
            # All affine equality constraints are independent
            return cons, 0

        # Keep only the first 'rank' rows (those corresponding to
        # linearly independent constraints)
        # P gives the permutation - P[:rank] are the indices of independent rows
        keep_row_indices = set(P[:rank])

        # Map back to original constraint indices
        keep_constraint_indices = set()
        for row_idx in keep_row_indices:
            keep_constraint_indices.add(eq_indices[row_idx])

        # Build filtered constraint list: independent affine eq + all others
        filtered_affine_eq = [c for orig_idx, c in affine_eq_cons if orig_idx in keep_constraint_indices]

        n_removed = len(affine_eq_cons) - len(filtered_affine_eq)
        if n_removed > 0 and verbose:
            logger.info(f"Removed {n_removed} redundant affine equality constraint(s)")

        return filtered_affine_eq + other_cons, n_removed

    except Exception as e:
        logger.debug(f"Constraint redundancy check failed: {e}")
        return cons, 0




from __future__ import annotations

import autograd.numpy as np

from ..expression import Expr, BaseExpr, ExprLike
from ..overrides import svd
from ..constants import Curvature as C


class PolarDecomposition(Expr):
    """Polar decomposition of a matrix.

    Returns the orthogonal factor R from X = R @ P where R is orthogonal
    and P is positive semidefinite. Uses SVD: R = U @ diag(S) @ Vt with
    sign correction to ensure det(R) > 0.

    Curvature is unknown (non-convex operation).
    """

    def __init__(self, left: ExprLike) -> None:
        super().__init__("polar_decomp", left)
        # Validate that input is 2D
        if isinstance(left, BaseExpr):
            if len(left.shape) != 2:
                raise ValueError("PolarDecomposition requires a 2D matrix input")
        elif hasattr(left, 'shape'):
            if len(left.shape) != 2:
                raise ValueError("PolarDecomposition requires a 2D matrix input")

    def __call__(self, x: np.ndarray) -> np.ndarray:
        U, _, Vt = svd(x, full_matrices=False)
        det_UVt = np.linalg.det(U @ Vt)
        S = np.ones((x.shape[0],))
        S[-1] = np.sign(det_UVt)
        return U @ np.diag(S) @ Vt

    @property
    def curvature(self) -> C:
        return C.UNKNOWN

    @property
    def shape(self) -> tuple[int, ...]:
        if isinstance(self.left, BaseExpr):
            return self.left.shape
        return np.shape(self.left) or (1,)


__all__ = [
    "det",
    "logdet",
    "norm",
    "PolarDecomposition",
    "sum",
    "trace",
    "maximum",
    "minimum",
    "amin",
    "amax",
    "abs",
    "log",
    "exp",
    "sqrt",
    "sin",
    "cos",
]

from .det import det
from .logdet import logdet
from .norm import norm
from .polar import PolarDecomposition
from .sum import sum
from .trace import trace
from .maximum import maximum
from .minimum import minimum
from .amin import amin
from .amax import amax
from .abs import abs
from .log import log
from .exp import exp
from .sqrt import sqrt
from .sin import sin
from .cos import cos



from __future__ import annotations

import autograd.numpy as np

from ..expression import Expr, BaseExpr, ExprLike
from ..constants import Curvature as C


class cos(Expr):
    """Elementwise cosine."""

    def __init__(self, left: ExprLike) -> None:
        super().__init__("cos", left)

    def __call__(self, x):
        return np.cos(x)

    @property
    def curvature(self) -> C:
        # Cosine is non-convex in general
        return C.UNKNOWN

    @property
    def shape(self) -> tuple[int, ...]:
        if isinstance(self.left, BaseExpr):
            return self.left.shape
        return np.shape(self.left) or (1,)



from __future__ import annotations

import autograd.numpy as np

from ..expression import Expr, BaseExpr, broadcast_shapes, ExprLike
from ..constants import Curvature as C


class minimum(Expr):
    """Element-wise minimum of two arrays.

    minimum(x, y) is concave (pointwise min of concave functions is concave).

    Curvature rules:
    - minimum(constant, constant) -> constant
    - minimum(affine, affine) -> concave
    - minimum(concave, concave) -> concave
    - minimum(convex, _) -> unknown
    """

    def __init__(self, left: ExprLike, right: ExprLike) -> None:
        super().__init__("minimum", left, right)

    def __call__(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        return np.minimum(x, y)

    @property
    def curvature(self) -> C:
        if isinstance(self.left, BaseExpr):
            left = self.left.curvature
        else:
            left = C.CONSTANT

        if isinstance(self.right, BaseExpr):
            right = self.right.curvature
        else:
            right = C.CONSTANT

        if left == C.CONSTANT and right == C.CONSTANT:
            return C.CONSTANT

        if left == C.CONSTANT:
            left = C.AFFINE
        if right == C.CONSTANT:
            right = C.AFFINE

        if left in (C.CONCAVE, C.AFFINE) and right in (C.CONCAVE, C.AFFINE):
            return C.CONCAVE
        return C.UNKNOWN

    @property
    def shape(self) -> tuple[int, ...]:
        left_shape = self.left.shape if isinstance(self.left, BaseExpr) else np.shape(self.left)
        right_shape = self.right.shape if isinstance(self.right, BaseExpr) else np.shape(self.right)
        return broadcast_shapes(left_shape, right_shape)



__all__ = [
    "Variable",
    "Constraint",
    "Set",
    "Expr",
    "Problem",
    "Maximize",
    "Minimize",
    # Gradient-free solvers
    "NELDER_MEAD",
    "POWELL",
    "COBYLA",
    "COBYQA",
    # Gradient-based solvers
    "CG",
    "BFGS",
    "LBFGSB",
    "TNC",
    "SLSQP",
    # Hessian-based solvers
    "NEWTON_CG",
    "DOGLEG",
    "TRUST_NCG",
    "TRUST_KRYLOV",
    "TRUST_EXACT",
    "TRUST_CONSTR",
    # Global optimizers
    "DIFF_EVOLUTION",
    "DUAL_ANNEALING",
    "SHGO",
    "BASINHOPPING",
    # Other
    "IPOPT",
    "BNB",
    "CONSTANT",
    "AFFINE",
    "CONVEX",
    "CONCAVE",
    "UNKNOWN",
    "det",
    "norm",
    "sum",
    "trace",
    "maximum",
    "minimum",
    "amin",
    "amax",
    "abs",
    "log",
    "exp",
    "sqrt",
    "sin",
    "cos",
    "logdet",
    "PolarDecomposition",
    "Function",
    "function",
    "Graph",
    "DiGraph",
    "SO",
    "PerspectiveCone",
    "DiscreteSet",
    "DiscreteRanges",
    "SolverStatus",
]

from .variable import Variable, reset_variable_ids as reset_variable_ids
from .constraint import Constraint
from .set import Set
from .expression import Expr
from .problem import Problem, Maximize, Minimize
from .constants import Solver, Curvature
from .solvers import SolverStatus

# Gradient-free solvers
NELDER_MEAD = Solver.NELDER_MEAD
POWELL = Solver.POWELL
COBYLA = Solver.COBYLA
COBYQA = Solver.COBYQA

# Gradient-based solvers
CG = Solver.CG
BFGS = Solver.BFGS
LBFGSB = Solver.LBFGSB
TNC = Solver.TNC
SLSQP = Solver.SLSQP

# Hessian-based solvers
NEWTON_CG = Solver.NEWTON_CG
DOGLEG = Solver.DOGLEG
TRUST_NCG = Solver.TRUST_NCG
TRUST_KRYLOV = Solver.TRUST_KRYLOV
TRUST_EXACT = Solver.TRUST_EXACT
TRUST_CONSTR = Solver.TRUST_CONSTR

# Global optimizers
DIFF_EVOLUTION = Solver.DIFF_EVOLUTION
DUAL_ANNEALING = Solver.DUAL_ANNEALING
SHGO = Solver.SHGO
BASINHOPPING = Solver.BASINHOPPING

# Other
IPOPT = Solver.IPOPT
BNB = Solver.BNB

CONSTANT = Curvature.CONSTANT
AFFINE = Curvature.AFFINE
CONVEX = Curvature.CONVEX
CONCAVE = Curvature.CONCAVE
UNKNOWN = Curvature.UNKNOWN

from .atoms import det, norm, sum, trace, maximum, minimum, amin, amax, abs, log, exp, sqrt, logdet, PolarDecomposition
from .atoms import sin, cos
from .constructs import Function, function, Graph, DiGraph
from .sets import SO, PerspectiveCone, DiscreteSet, DiscreteRanges



from __future__ import annotations

from typing import Dict

from ..constants import Solver
from .base import (
    ConstraintData,
    ProblemData,
    SolverBackend,
    SolverResult,
    SolverStats,
    SolverStatus,
)
from .scipy_backend import ScipyBackend
from .ipopt_backend import IpoptBackend
from .bnb import BranchAndBoundBackend
from .global_scipy_backend import GlobalScipyBackend


_SCIPY_BACKEND = ScipyBackend()
_IPOPT_BACKEND = IpoptBackend()
_BNB_BACKEND = BranchAndBoundBackend()
_GLOBAL_SCIPY_BACKEND = GlobalScipyBackend()


_SOLVER_BACKENDS: Dict[str, SolverBackend] = {
    # Gradient-free
    Solver.NELDER_MEAD.value: _SCIPY_BACKEND,
    Solver.POWELL.value: _SCIPY_BACKEND,
    Solver.COBYLA.value: _SCIPY_BACKEND,
    Solver.COBYQA.value: _SCIPY_BACKEND,
    # Gradient-based
    Solver.CG.value: _SCIPY_BACKEND,
    Solver.BFGS.value: _SCIPY_BACKEND,
    Solver.LBFGSB.value: _SCIPY_BACKEND,
    Solver.TNC.value: _SCIPY_BACKEND,
    Solver.SLSQP.value: _SCIPY_BACKEND,
    # Hessian-based
    Solver.NEWTON_CG.value: _SCIPY_BACKEND,
    Solver.DOGLEG.value: _SCIPY_BACKEND,
    Solver.TRUST_NCG.value: _SCIPY_BACKEND,
    Solver.TRUST_KRYLOV.value: _SCIPY_BACKEND,
    Solver.TRUST_EXACT.value: _SCIPY_BACKEND,
    Solver.TRUST_CONSTR.value: _SCIPY_BACKEND,
    # Global optimizers
    Solver.DIFF_EVOLUTION.value: _GLOBAL_SCIPY_BACKEND,
    Solver.DUAL_ANNEALING.value: _GLOBAL_SCIPY_BACKEND,
    Solver.SHGO.value: _GLOBAL_SCIPY_BACKEND,
    Solver.BASINHOPPING.value: _GLOBAL_SCIPY_BACKEND,
    # Other
    Solver.IPOPT.value: _IPOPT_BACKEND,
    Solver.BNB.value: _BNB_BACKEND,
}


def register_solver_backend(solver_name: str, backend: SolverBackend) -> None:
    _SOLVER_BACKENDS[solver_name] = backend


def get_solver_backend(solver: Solver | str) -> SolverBackend:
    solver_name = solver.value if isinstance(solver, Solver) else str(solver)
    if solver_name not in _SOLVER_BACKENDS:
        raise ValueError(f"No solver backend registered for solver '{solver_name}'")
    return _SOLVER_BACKENDS[solver_name]


__all__ = [
    "ConstraintData",
    "ProblemData",
    "SolverBackend",
    "SolverResult",
    "SolverStats",
    "SolverStatus",
    "IpoptBackend",
    "get_solver_backend",
    "register_solver_backend",
]




from __future__ import annotations

import time
from typing import Dict

import autograd.numpy as np

from .constants import Solver, DEFAULT_PROJECTION_TOL
from .parser import collect_vars
from .sets.discrete_set import DiscreteSet, DiscreteRanges
from .solvers import ProblemData, get_solver_backend


class Minimize:
    def __init__(self, expr):
        self.expr = expr


class Maximize:
    def __init__(self, expr):
        self.expr = -expr


class Problem:
    """An optimization problem with objective and constraints."""

    def __init__(self, objective: Minimize | Maximize, constraints=None):
        if not isinstance(objective, (Minimize, Maximize)):
            raise TypeError(
                f"Objective must be Minimize or Maximize, got {type(objective).__name__}"
            )
        self.objective = objective
        self.constraints = list(constraints) if constraints is not None else []

        self.vars = []
        self.var_shapes = {}
        self.var_slices = {}
        self.total_size = 0

        all_vars = []
        collect_vars(objective.expr, all_vars)
        for c in self.constraints:
            collect_vars(c.left, all_vars)
            collect_vars(c.right, all_vars)

        # Deduplicate variables before collecting their embedded constraints
        self.vars = []
        for v in all_vars:
            if v.name not in self.var_shapes:
                self.var_shapes[v.name] = v.shape
                self.var_slices[v.name] = (self.total_size, self.total_size + v.size)
                self.total_size += v.size
                self.vars.append(v)

        # Now collect embedded constraints from unique variables only
        variable_constraints = []
        for v in self.vars:
            variable_constraints.extend(v.constraints)
        self.constraints += variable_constraints

        self.status = None
        self.solver_stats = None

    def _has_discrete_constraints(self, constraints) -> bool:
        """Check if any constraint involves a DiscreteSet or DiscreteRanges."""
        for c in constraints:
            if c.op == "in" and isinstance(c.right, (DiscreteSet, DiscreteRanges)):
                return True
        return False

    def _select_default_solver(self, has_integers: bool, has_discrete: bool, has_constraints: bool) -> Solver:
        """Select the default solver based on problem characteristics."""
        if has_integers or has_discrete:
            return Solver.BNB
        if has_constraints:
            return Solver.SLSQP
        return Solver.LBFGSB

    def solve(self, solver=None, solver_options=None, presolve=False, compile=False):
        """
        Solve the optimization problem.

        Args:
            solver: The solver to use. If None, automatically selects:
                    - BNB for problems with integer variables or discrete constraints
                    - SLSQP for problems with constraints
                    - L-BFGS-B for unconstrained problems
            solver_options: Options to pass to the solver
            presolve: Whether to run a presolve phase
            compile: If True, compile expressions for faster evaluation.
                     This can significantly speed up problems with complex
                     expressions that are evaluated many times.

        Returns:
            SolverResult with solution and status
        """
        solver_options = solver_options or {}

        problem_data, backend_options = self._compile_problem_data(
            solver_options, presolve, compile
        )

        # Auto-select solver if not specified
        if solver is None:
            has_integers = bool(problem_data.integer_vars)
            has_discrete = self._has_discrete_constraints(problem_data.constraints)
            has_constraints = bool(problem_data.constraints)
            solver = self._select_default_solver(has_integers, has_discrete, has_constraints)

        backend = get_solver_backend(solver)
        solver_name = solver.value if isinstance(solver, Solver) else str(solver)

        has_discrete = self._has_discrete_constraints(problem_data.constraints)
        if (problem_data.integer_vars or has_discrete) and solver_name != Solver.BNB.value:
            raise ValueError(
                "Integer variables or discrete constraints detected; use the BnB solver (nvx.BNB)."
            )

        result = backend.solve(problem_data, solver_name, backend_options)

        self.status = result.status
        self.solver_stats = result.stats

        sol_vars = problem_data.unpack(result.x)

        for v in self.vars:
            v.value = sol_vars[v.name]

        return result

    def _compile_problem_data(
        self, solver_options: Dict[str, object], presolve: bool, compile: bool
    ) -> tuple[ProblemData, Dict[str, object]]:
        options = dict(solver_options)

        p_tol = options.pop("p_tol", DEFAULT_PROJECTION_TOL)
        p_maxiter = options.pop("p_maxiter", 100)

        if p_tol <= 0:
            raise ValueError(f"p_tol must be positive, got {p_tol}")
        if p_maxiter <= 0:
            raise ValueError(f"p_maxiter must be positive, got {p_maxiter}")

        start_setup_time = time.time()

        x0 = np.ones(self.total_size)
        var_names = [v.name for v in self.vars]
        var_shapes = dict(self.var_shapes)
        var_slices = dict(self.var_slices)
        integer_vars = tuple(v.name for v in self.vars if getattr(v, "is_integer", False))

        for v in self.vars:
            v_start, v_end = self.var_slices[v.name]
            if v.value is None:
                x0[v_start:v_end] = np.zeros(v.size)
            else:
                x0[v_start:v_end] = np.ravel(v.value)

        setup_time = time.time() - start_setup_time

        problem_data = ProblemData(
            x0=x0,
            var_names=var_names,
            var_shapes=var_shapes,
            var_slices=var_slices,
            objective_expr=self.objective.expr,
            constraints=tuple(self.constraints),
            integer_vars=integer_vars,
            projection_tolerance=p_tol,
            projection_maxiter=p_maxiter,
            presolve=presolve,
            compile=compile,
            setup_time=setup_time,
        )

        return problem_data, options



from __future__ import annotations

import autograd.numpy as np

from ..expression import Expr, BaseExpr, ExprLike
from ..constants import Curvature as C


class sum(Expr):
    """Sum of array elements along an axis.

    sum(x) returns the sum of all elements (scalar if axis=None).
    Preserves curvature (affine operation).
    """

    def __init__(self, left: ExprLike, axis: int | None = None) -> None:
        self.axis = axis
        super().__init__("sum", left)

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return np.sum(x, axis=self.axis)

    @property
    def curvature(self) -> C:
        arg = self.left
        arg_curv = arg.curvature if isinstance(arg, BaseExpr) else C.CONSTANT

        return arg_curv

    @property
    def shape(self) -> tuple[int, ...]:
        if self.axis is None:
            return (1,)
        arg_shape = self.left.shape if isinstance(self.left, BaseExpr) else np.shape(self.left)
        ndim = len(arg_shape)
        axis = self.axis if self.axis >= 0 else ndim + self.axis
        if axis < 0 or axis >= ndim:
            raise ValueError(f"axis {self.axis} is out of bounds for array of dimension {ndim}")
        result = list(arg_shape)
        result.pop(axis)
        return tuple(result) if result else (1,)


from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence, Union, Tuple, List

import autograd.numpy as np

from ..set import Set
from ..constraint import Constraint
from ..constants import DEFAULT_DISCRETE_TOL


@dataclass(frozen=True)
class Range:
    """A continuous range [lb, ub] within a MixedDiscreteSet."""
    lb: float
    ub: float

    def __post_init__(self):
        if self.lb > self.ub:
            raise ValueError(f"Range lower bound {self.lb} > upper bound {self.ub}")

    def __contains__(self, value: float) -> bool:
        return self.lb <= value <= self.ub

    def __repr__(self):
        return f"[{self.lb}, {self.ub}]"


class DiscreteSet(Set):
    """
    A set of allowed discrete values (integers or floats).

    Used to constrain a variable to take values from a specific discrete set.

    Example:
        x = Variable(integer=True)
        constraint = x ^ DiscreteSet([1, 10, 100, 9, -2])
        # or using the list shorthand:
        constraint = x ^ [1, 10, 100, 9, -2]

        # Also works with floats:
        y = Variable()
        constraint = y ^ [0.1, 0.5, 1.0, 2.5]
    """

    def __init__(self, values: Sequence[Union[int, float]], tolerance: float = DEFAULT_DISCRETE_TOL):
        """
        Create a DiscreteSet with allowed values.

        Args:
            values: A sequence of allowed values (int or float)
            tolerance: Tolerance for checking membership (default: DEFAULT_DISCRETE_TOL)
        """
        # Keep original values, sort them, remove duplicates within tolerance
        sorted_vals = sorted(set(float(v) for v in values))
        # Remove duplicates that are too close
        unique_vals = []
        for v in sorted_vals:
            if not unique_vals or not np.isclose(v, unique_vals[-1], atol=tolerance):
                unique_vals.append(v)

        self._values = tuple(unique_vals)
        self._tolerance = tolerance
        if not self._values:
            raise ValueError("DiscreteSet must contain at least one value")
        super().__init__(name=f"DiscreteSet({list(self._values)})")

    @property
    def values(self) -> tuple:
        """The sorted tuple of allowed values."""
        return self._values

    @property
    def tolerance(self) -> float:
        """The tolerance for membership checking."""
        return self._tolerance

    def constrain(self, var):
        """
        Create a discrete membership constraint.

        Args:
            var: The variable to constrain

        Returns:
            Constraint object with "in" operator

        Raises:
            ValueError: If var is an integer variable but set contains non-integers
        """
        # Check if variable is integer-constrained
        if getattr(var, "is_integer", False):
            non_integers = [v for v in self._values if v != int(v)]
            if non_integers:
                raise ValueError(
                    f"Cannot constrain integer variable to DiscreteSet containing "
                    f"non-integer values: {non_integers}. Either declare the variable "
                    f"as continuous or use integer values in the set."
                )
        return Constraint(var, "in", self)

    def __contains__(self, value) -> bool:
        """Check if a value is in the set (within tolerance)."""
        return any(np.isclose(float(value), v, atol=self._tolerance) for v in self._values)

    def nearest(self, value: float) -> float:
        """Find the nearest value in the set."""
        return min(self._values, key=lambda v: abs(v - value))

    def values_below(self, value: float) -> tuple:
        """Return all values in the set strictly below the given value."""
        return tuple(v for v in self._values if v < value - self._tolerance)

    def values_above(self, value: float) -> tuple:
        """Return all values in the set strictly above the given value."""
        return tuple(v for v in self._values if v > value + self._tolerance)

    def __len__(self) -> int:
        return len(self._values)

    def __iter__(self):
        return iter(self._values)

    def __repr__(self):
        return f"DiscreteSet({list(self._values)})"


class DiscreteRanges(Set):
    """
    A set of disjoint continuous ranges (intervals).

    Constrains a variable to be within one of several continuous ranges.
    The variable is treated as continuous once committed to a specific range.
    Useful for modeling disjunctive constraints without Big-M formulations.

    Example:
        # Time slots: 9am-12pm or 1pm-5pm
        slots = [[9, 12], [13, 17]]
        constraint = start_time ^ slots

        # Temperature zones
        zones = [[0, 10], [25, 35], [50, 60]]
        constraint = temp ^ zones
    """

    def __init__(
        self,
        ranges: Sequence[Sequence],
        tolerance: float = DEFAULT_DISCRETE_TOL
    ):
        """
        Create a DiscreteRanges with continuous ranges.

        Args:
            ranges: A sequence of 2-element [lb, ub] ranges
            tolerance: Tolerance for merging adjacent ranges
        """
        parsed_ranges: List[Range] = []

        for item in ranges:
            if isinstance(item, (list, tuple)) and len(item) == 2:
                lb, ub = float(item[0]), float(item[1])
                parsed_ranges.append(Range(lb, ub))
            else:
                raise ValueError(
                    f"Invalid range {item}: must be a 2-element [lb, ub] sequence"
                )

        if not parsed_ranges:
            raise ValueError("DiscreteRanges must contain at least one range")

        # Sort ranges by lower bound
        parsed_ranges = sorted(parsed_ranges, key=lambda r: r.lb)

        # Merge overlapping ranges
        merged_ranges: List[Range] = []
        for r in parsed_ranges:
            if merged_ranges and r.lb <= merged_ranges[-1].ub + tolerance:
                # Overlapping or adjacent, merge
                merged_ranges[-1] = Range(merged_ranges[-1].lb, max(merged_ranges[-1].ub, r.ub))
            else:
                merged_ranges.append(r)

        self._ranges = tuple(merged_ranges)
        self._tolerance = tolerance

        # Build name
        parts = [repr(r) for r in self._ranges]
        super().__init__(name=f"DiscreteRanges({parts})")

    @property
    def ranges(self) -> Tuple[Range, ...]:
        """The sorted tuple of continuous ranges."""
        return self._ranges

    @property
    def tolerance(self) -> float:
        """The tolerance for membership checking."""
        return self._tolerance

    @property
    def num_branches(self) -> int:
        """Number of disjuncts (ranges)."""
        return len(self._ranges)

    def constrain(self, var):
        """Create a discrete ranges membership constraint."""
        return Constraint(var, "in", self)

    def __contains__(self, value: float) -> bool:
        """Check if a value is in any range."""
        return any(value in r for r in self._ranges)

    def nearest(self, value: float) -> float:
        """Find the nearest value in any range."""
        candidates = []

        for r in self._ranges:
            if value in r:
                candidates.append((0.0, value))
            else:
                dist_lb = abs(value - r.lb)
                dist_ub = abs(value - r.ub)
                if dist_lb < dist_ub:
                    candidates.append((dist_lb, r.lb))
                else:
                    candidates.append((dist_ub, r.ub))

        return min(candidates, key=lambda x: x[0])[1]

    def ranges_below(self, value: float) -> Tuple[Range, ...]:
        """Return ranges strictly below the given value."""
        ranges_below = []
        for r in self._ranges:
            if r.ub < value - self._tolerance:
                ranges_below.append(r)
            elif r.lb < value - self._tolerance:
                # Range straddles value, truncate it
                ranges_below.append(Range(r.lb, value - self._tolerance))
        return tuple(ranges_below)

    def ranges_above(self, value: float) -> Tuple[Range, ...]:
        """Return ranges strictly above the given value."""
        ranges_above = []
        for r in self._ranges:
            if r.lb > value + self._tolerance:
                ranges_above.append(r)
            elif r.ub > value + self._tolerance:
                # Range straddles value, truncate it
                ranges_above.append(Range(value + self._tolerance, r.ub))
        return tuple(ranges_above)

    def bounds(self) -> Tuple[float, float]:
        """Get the overall bounds of the set."""
        return self._ranges[0].lb, self._ranges[-1].ub

    def __repr__(self):
        return f"DiscreteRanges({[[r.lb, r.ub] for r in self._ranges]})"


def _coerce_to_discrete_set(obj) -> Union[DiscreteSet, DiscreteRanges]:
    """
    Coerce a list/tuple to DiscreteSet or DiscreteRanges if needed.

    This helper allows the shorthand:
        x ^ [1, 2, 3]              -> DiscreteSet (all scalars)
        x ^ [[0, 5], [10, 15]]     -> DiscreteRanges (all ranges)

    Note: Mixed discrete values and ranges are NOT supported.
    Use either all scalars (DiscreteSet) or all ranges (DiscreteRanges).
    """
    if isinstance(obj, (DiscreteSet, DiscreteRanges)):
        return obj
    if isinstance(obj, (list, tuple)):
        # Check if items are ranges or scalars
        has_range = any(
            isinstance(item, (list, tuple)) and len(item) == 2
            for item in obj
        )
        has_scalar = any(
            isinstance(item, (int, float))
            for item in obj
        )

        if has_range and has_scalar:
            raise ValueError(
                "Cannot mix discrete values and ranges. "
                "Use DiscreteSet for discrete values [1, 2, 3] or "
                "DiscreteRanges for ranges [[0, 5], [10, 15]]."
            )

        if has_range:
            return DiscreteRanges(obj)
        return DiscreteSet(obj)
    raise TypeError(f"Cannot convert {type(obj)} to DiscreteSet or DiscreteRanges")



"""
Outer Approximation (OA) Cut Generation and Management

This module handles the generation and management of outer approximation cuts
for the branch-and-bound MINLP solver.

OA cuts are used to tighten the NLP relaxation at each node for convex problems.
For convex constraints g(x) >= 0, the linearization is:
    g(x*) + ∇g(x*)ᵀ(x - x*) >= 0
    => ∇g(x*)ᵀ x >= ∇g(x*)ᵀ x* - g(x*)

WARNING: For non-convex problems, OA cuts can cut off the global optimum!
Only use OA cuts if you are certain the problem is convex.
"""

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Dict, List

import autograd.numpy as np

logger = logging.getLogger(__name__)


@dataclass
class OACut:
    """
    An outer approximation cut (linear constraint).

    For constraint cuts: a^T x >= b (linearization of g(x) >= 0)

    For convex constraints g(x) >= 0, the linearization is:
        g(x*) + ∇g(x*)ᵀ(x - x*) >= 0
        => ∇g(x*)ᵀ x >= ∇g(x*)ᵀ x* - g(x*)

    Note: Epigraph cuts for the objective (t >= f(x)) are NOT generated here.
    They would require an explicit epigraph variable t in the formulation.
    The NLP relaxations minimize f(x) directly without epigraph reformulation.
    """
    coefficients: np.ndarray  # a (gradient coefficients for x)
    rhs: float                # b (right-hand side)
    is_equality: bool = False
    age: int = 0              # Number of nodes since cut was added
    times_active: int = 0     # Number of times cut was binding


def generate_oa_cuts(x: np.ndarray, cons: List[Dict]) -> List[OACut]:
    """
    Generate outer approximation cuts at current point.

    For convex constraints g(x) >= 0, the linearization is:
        g(x*) + ∇g(x*)ᵀ(x - x*) >= 0
        => ∇g(x*)ᵀ x >= ∇g(x*)ᵀ x* - g(x*)

    Implementation notes:
    - Only inequality constraints are used as OA cuts in _solve_node_nlp.
      Equality constraint linearizations are generated but filtered out
      when adding cuts to the NLP (is_equality=True cuts are skipped).
    - Epigraph cuts for the objective would require an explicit epigraph
      variable t in the formulation (min t s.t. t >= f(x)). Since our NLP
      relaxations minimize f(x) directly, epigraph cuts are not applicable.

    Args:
        x: Current solution point
        cons: List of scipy constraint dictionaries

    Returns:
        List of OACut objects (inequality cuts are used, equality cuts stored
        but not applied in node NLPs)
    """
    cuts = []

    for con in cons:
        try:
            con_val = con["fun"](x)
            con_jac = con["jac"](x)

            if con_val.ndim == 0:
                con_val = np.array([con_val])
            if con_jac.ndim == 1:
                con_jac = con_jac.reshape(1, -1)

            for i in range(len(con_val)):
                if np.all(np.isfinite(con_jac[i])):
                    grad_dot_x = np.dot(con_jac[i], x)
                    rhs = float(grad_dot_x - con_val[i])
                    cuts.append(OACut(
                        coefficients=con_jac[i].copy(),
                        rhs=rhs,
                        is_equality=(con["type"] == "eq"),
                    ))
        except Exception as e:
            logger.debug(f"Constraint cut generation failed: {e}")

    return cuts


def prune_cut_pool(
    cuts: List[OACut],
    max_cuts: int,
    max_age: int,
) -> List[OACut]:
    """
    Prune cut pool to remove old/inactive cuts.
    
    Args:
        cuts: List of OA cuts
        max_cuts: Maximum number of cuts to keep
        max_age: Maximum age before considering removal
        
    Returns:
        Filtered list of cuts
    """
    if len(cuts) <= max_cuts:
        # Just remove very old inactive cuts
        return [c for c in cuts if c.age < max_age or c.times_active > 0]

    # Score cuts: newer and more active cuts are better
    # Keep cuts that are either recent or frequently active
    def cut_score(cut: OACut) -> float:
        # Higher score = better cut to keep
        recency = max(0, max_age - cut.age) / max_age
        activity = min(cut.times_active, 10) / 10  # Cap activity score
        return recency * 0.4 + activity * 0.6

    scored = [(cut_score(c), i, c) for i, c in enumerate(cuts)]
    scored.sort(reverse=True)

    return [c for _, _, c in scored[:max_cuts]]




from __future__ import annotations

from dataclasses import dataclass
from enum import StrEnum
from typing import Any, Callable, Dict, List, Protocol, Sequence, Tuple

import autograd.numpy as anp  # type: ignore


ArrayLike = anp.ndarray


class SolverStatus(StrEnum):
    OPTIMAL = "optimal"
    SUBOPTIMAL = "suboptimal"
    INFEASIBLE = "infeasible"
    UNBOUNDED = "unbounded"
    MAX_ITERATIONS = "max_iterations"
    NUMERICAL_ERROR = "numerical_error"
    ERROR = "error"
    UNKNOWN = "unknown"


@dataclass
class ConstraintData:
    type: str
    fun: Callable[[ArrayLike], ArrayLike]
    jac: Callable[[ArrayLike], ArrayLike]
    op: str


@dataclass
class ProblemData:
    x0: ArrayLike
    var_names: List[str]
    var_shapes: Dict[str, Tuple[int, ...]]
    var_slices: Dict[str, Tuple[int, int]]
    objective_expr: Any
    constraints: Sequence[Any]
    integer_vars: Sequence[str]
    projection_tolerance: float
    projection_maxiter: int
    presolve: bool
    compile: bool = False
    setup_time: float = 0.0

    def unpack(self, x: Sequence[float]) -> Dict[str, ArrayLike]:
        if isinstance(x, anp.ndarray):
            arr = x
        else:
            arr = anp.array(x)
        var_dict: Dict[str, ArrayLike] = {}
        for name in self.var_names:
            start, end = self.var_slices[name]
            shape = self.var_shapes[name]
            segment = arr[start:end]
            if shape:
                var_dict[name] = segment.reshape(shape)
            else:
                var_dict[name] = segment[0] if segment.size == 1 else segment
        return var_dict


@dataclass
class SolverStats:
    solver_name: str
    solve_time: float | None = None
    setup_time: float | None = None
    num_iters: int | None = None


@dataclass
class SolverResult:
    x: ArrayLike
    status: SolverStatus
    stats: SolverStats
    raw_result: object | None = None


class SolverBackend(Protocol):
    def solve(
        self,
        problem_data: ProblemData,
        solver: str,
        solver_options: Dict[str, object],
    ) -> SolverResult:
        ...


# =============================================================================
# Shared Solver Utilities
# =============================================================================

def build_objective(problem_data: ProblemData) -> Callable[[ArrayLike], float]:
    """
    Build an objective function from problem data.
    
    This is shared across all solver backends. When `problem_data.compile` is True,
    uses the codegen compiler for better performance. Otherwise uses the interpreter.
    
    Args:
        problem_data: The problem specification containing the objective expression
        
    Returns:
        A callable that takes a flat variable array and returns the objective value
    """
    from ..parser import eval_expression
    from ..compiler import compile_to_function
    
    if problem_data.compile:
        compiled_obj = compile_to_function(problem_data.objective_expr)

        def obj(x):
            var_dict = problem_data.unpack(x)
            return compiled_obj(var_dict)
    else:
        def obj(x):
            var_dict = problem_data.unpack(x)
            return eval_expression(problem_data.objective_expr, var_dict)

    return obj


def uses_projection(problem_data: ProblemData) -> bool:
    """
    Check if any constraint in the problem requires projection.
    
    Projection constraints use the "<-" operator and require a two-phase solve:
    first the main optimization, then a projection step to satisfy the constraint.
    
    Args:
        problem_data: The problem specification
        
    Returns:
        True if any constraint uses the "<-" projection operator
    """
    return any(getattr(c, "op", None) == "<-" for c in problem_data.constraints)





from ..set import Set
from ..constraint import Constraint
from ..atoms.polar import PolarDecomposition


class SO(Set):
    def __init__(self, n):
        super().__init__(name=f"SO({n})")
        self.n = n

    def constrain(self, var):
        if var.shape != (self.n, self.n):
            raise ValueError(f"Variable shape {var.shape} does not match SO({self.n}), expected ({self.n}, {self.n})")
        return Constraint(var, "<-", PolarDecomposition(var))



"""
Expression Compiler for nvxpy

This module provides compiled evaluation approaches for expression trees.
Instead of recursively interpreting the tree at each evaluation, we compile
the expression once and execute it efficiently.

Two Compilation Strategies
==========================

1. **Interpreted Compilation** (CompiledExpression / compile_expression)
   
   Compiles the expression tree into a sequence of operations (linear IR)
   that are executed by a Python interpreter loop.
   
   Pros:
   - Easier to debug (can inspect the operation sequence)
   - More flexible for dynamic modifications
   - Lower compilation overhead
   
   Cons:
   - Slightly slower execution due to interpreter loop overhead
   
   Use when:
   - Debugging expression evaluation
   - Expressions are evaluated only a few times
   - You need to inspect the compiled operations

2. **Codegen Compilation** (CodegenCompiler / compile_to_function)
   
   Generates actual Python source code and compiles it using exec().
   This creates a native Python function with no interpreter overhead.
   
   Pros:
   - Fastest execution (native Python speed)
   - No per-operation dispatch overhead
   - Common subexpression elimination (CSE) built-in
   
   Cons:
   - Slightly higher compilation overhead
   - Harder to debug (generated code)
   
   Use when:
   - Expressions are evaluated many times (optimization loops)
   - Maximum performance is needed
   - Using the `compile=True` option in Problem.solve()

Usage Examples
==============

Interpreted compilation::

    from nvxpy.compiler import compile_expression
    
    x = Variable((3,), name="x")
    expr = x + 2 * x
    compiled = compile_expression(expr)
    result = compiled({"x": np.array([1, 2, 3])})

Codegen compilation::

    from nvxpy.compiler import compile_to_function
    
    x = Variable((3,), name="x")
    expr = x + 2 * x
    compiled_func = compile_to_function(expr)
    result = compiled_func({"x": np.array([1, 2, 3])})

Unified interface with strategy choice::

    from nvxpy.compiler import compile_expr
    
    # Use codegen (default, fastest)
    compiled = compile_expr(expr, strategy="codegen")
    
    # Use interpreted
    compiled = compile_expr(expr, strategy="interpreted")

Internal Architecture
=====================

The compilation process (for both strategies):
1. Traverse the expression tree once
2. Emit operations in topological order  
3. Store constants and operation metadata
4. Return a compiled function that executes the operations

This eliminates:
- Recursive function call overhead
- Repeated isinstance() checks
- String comparisons for operation dispatch
"""

from typing import Callable, Dict, List, Tuple, Any
from dataclasses import dataclass, field
import autograd.numpy as np

from .expression import Expr
from .variable import Variable
from .constructs.function import Function


@dataclass
class CompiledOp:
    """A single compiled operation."""
    op: str                          # Operation name
    result_idx: int                  # Index to store result
    left_idx: int                    # Index of left operand (-1 for constants)
    right_idx: int                   # Index of right operand (-1 for None/constants)
    left_const: Any = None           # Left constant value (if left_idx == -1)
    right_const: Any = None          # Right constant value (if right_idx == -1)
    atom_callable: Callable = None   # For atom operations (norm, trace, etc.)
    func_callable: Callable = None   # For Function operations
    func_kwargs: Dict = None         # Kwargs for Function
    getitem_key: Any = None          # For getitem operations
    # For container operations (dict, list, tuple)
    elem_indices: List[int] = None   # Slot indices of elements (-1 for constants)
    elem_consts: List[Any] = None    # Constant values for elements
    dict_keys: List[Any] = None      # Keys for dict elements
    container_type: type = None      # list or tuple (for preserving type)


@dataclass
class CompiledExpression:
    """A compiled expression ready for fast evaluation."""
    ops: List[CompiledOp]            # Sequence of operations
    var_indices: Dict[str, int]      # Map variable names to slot indices
    result_idx: int                  # Index of final result
    num_slots: int                   # Total number of temporary slots needed

    # Pre-allocated workspace (created on first eval, reused)
    _workspace: List | None = field(default=None, repr=False)

    def __call__(self, var_dict: Dict[str, np.ndarray]) -> np.ndarray:
        """Evaluate the compiled expression with given variable values."""
        return eval_compiled(self, var_dict)


# Operation dispatch table - maps op names to functions
# Using a dict avoids if-elif chains
_BINARY_OPS = {
    "add": lambda left, right: left + right,
    "sub": lambda left, right: left - right,
    "mul": lambda left, right: left * right,
    "div": lambda left, right: left / right,
    "pow": lambda left, right: left ** right,
    "matmul": lambda left, right: left @ right,
}

_UNARY_OPS = {
    "neg": lambda val: -val,
    "transpose": lambda val: val.T,
    "flatten": lambda val: val.flatten(),
}


class ExpressionCompiler:
    """Compiles expression trees into linear operation sequences."""

    def __init__(self):
        self.ops: List[CompiledOp] = []
        self.slot_counter = 0
        self.expr_to_slot: Dict[int, int] = {}  # expr id -> slot index
        self.var_indices: Dict[str, int] = {}    # var name -> slot index
        self.var_objects: Dict[str, Variable] = {}  # var name -> Variable object

    def _alloc_slot(self) -> int:
        """Allocate a new slot for intermediate results."""
        idx = self.slot_counter
        self.slot_counter += 1
        return idx

    def compile(self, expr) -> CompiledExpression:
        """Compile an expression tree into a CompiledExpression."""
        self.ops = []
        self.slot_counter = 0
        self.expr_to_slot = {}
        self.var_indices = {}
        self.var_objects = {}

        result_idx = self._compile_node(expr)

        return CompiledExpression(
            ops=self.ops,
            var_indices=self.var_indices,
            result_idx=result_idx,
            num_slots=self.slot_counter,
        )

    def _compile_node(self, expr) -> int:
        """
        Compile a single node and return its slot index.
        Returns -1 for constants (stored inline in the op).
        """
        # Check if already compiled (enables CSE)
        expr_id = id(expr)
        if expr_id in self.expr_to_slot:
            return self.expr_to_slot[expr_id]

        if isinstance(expr, Variable):
            # Variables get their own slots, filled at eval time
            if expr.name not in self.var_indices:
                slot = self._alloc_slot()
                self.var_indices[expr.name] = slot
                self.var_objects[expr.name] = expr
            return self.var_indices[expr.name]

        elif isinstance(expr, Function):
            # Compile all function arguments
            arg_indices = []
            arg_consts = []
            for arg in expr.args:
                idx = self._compile_node(arg)
                arg_indices.append(idx)
                if idx == -1:
                    arg_consts.append(arg)
                else:
                    arg_consts.append(None)

            result_slot = self._alloc_slot()

            # Create a special op for functions
            op = CompiledOp(
                op="func",
                result_idx=result_slot,
                left_idx=arg_indices[0] if arg_indices else -1,
                right_idx=arg_indices[1] if len(arg_indices) > 1 else -1,
                left_const=arg_consts[0] if arg_consts else None,
                right_const=arg_consts[1] if len(arg_consts) > 1 else None,
                func_callable=expr.func,
                func_kwargs=expr.kwargs if hasattr(expr, 'kwargs') else {},
            )
            # Store extra args for functions with >2 arguments
            op._extra_arg_indices = arg_indices[2:] if len(arg_indices) > 2 else []
            op._extra_arg_consts = arg_consts[2:] if len(arg_consts) > 2 else []

            self.ops.append(op)
            self.expr_to_slot[expr_id] = result_slot
            return result_slot

        elif isinstance(expr, Expr):
            # Compile left operand
            left_idx = self._compile_node(expr.left)
            left_const = expr.left if left_idx == -1 else None

            # Compile right operand if present
            if expr.right is not None:
                right_idx = self._compile_node(expr.right)
                right_const = expr.right if right_idx == -1 else None
            else:
                right_idx = -1
                right_const = None

            result_slot = self._alloc_slot()

            # Check if this is an atom (callable Expr like norm, trace, etc.)
            atom_callable = None
            if callable(expr) and expr.op not in _BINARY_OPS and expr.op not in _UNARY_OPS:
                atom_callable = expr

            # Special handling for getitem - store the key
            getitem_key = None
            if expr.op == "getitem":
                getitem_key = expr.right
                right_idx = -1  # Key is not a compiled expression
                right_const = None

            op = CompiledOp(
                op=expr.op,
                result_idx=result_slot,
                left_idx=left_idx,
                right_idx=right_idx,
                left_const=left_const,
                right_const=right_const,
                atom_callable=atom_callable,
                getitem_key=getitem_key,
            )
            self.ops.append(op)
            self.expr_to_slot[expr_id] = result_slot
            return result_slot

        elif isinstance(expr, dict):
            # Compile dict - recursively compile each value
            elem_indices = []
            elem_consts = []
            dict_keys = []
            for k, v in expr.items():
                idx = self._compile_node(v)
                elem_indices.append(idx)
                elem_consts.append(v if idx == -1 else None)
                dict_keys.append(k)

            result_slot = self._alloc_slot()
            op = CompiledOp(
                op="dict",
                result_idx=result_slot,
                left_idx=-1,
                right_idx=-1,
                elem_indices=elem_indices,
                elem_consts=elem_consts,
                dict_keys=dict_keys,
            )
            self.ops.append(op)
            self.expr_to_slot[expr_id] = result_slot
            return result_slot

        elif isinstance(expr, (list, tuple)) and not isinstance(expr, np.ndarray):
            # Compile list/tuple - recursively compile each element
            elem_indices = []
            elem_consts = []
            for e in expr:
                idx = self._compile_node(e)
                elem_indices.append(idx)
                elem_consts.append(e if idx == -1 else None)

            result_slot = self._alloc_slot()
            op = CompiledOp(
                op="container",
                result_idx=result_slot,
                left_idx=-1,
                right_idx=-1,
                elem_indices=elem_indices,
                elem_consts=elem_consts,
                container_type=type(expr),
            )
            self.ops.append(op)
            self.expr_to_slot[expr_id] = result_slot
            return result_slot

        else:
            # Constants (int, float, ndarray) - return -1 to indicate inline
            return -1


def compile_expression(expr) -> CompiledExpression:
    """
    Compile an expression tree into a fast-evaluating CompiledExpression.

    Args:
        expr: An expression tree (Expr, Variable, Function, or constant)

    Returns:
        CompiledExpression that can be called with a var_dict

    Example:
        >>> x = Variable((3,), name="x")
        >>> y = Variable((3,), name="y")
        >>> expr = x + y * 2
        >>> compiled = compile_expression(expr)
        >>> result = compiled({"x": np.array([1,2,3]), "y": np.array([4,5,6])})
    """
    compiler = ExpressionCompiler()
    return compiler.compile(expr)


def eval_compiled(compiled: CompiledExpression, var_dict: Dict[str, np.ndarray]) -> np.ndarray:
    """
    Evaluate a compiled expression with the given variable values.

    This is the fast path - no recursion, no isinstance checks per node,
    just sequential operation execution.
    """
    # Allocate or reuse workspace
    workspace = [None] * compiled.num_slots

    # Load variables into their slots
    for name, slot in compiled.var_indices.items():
        workspace[slot] = var_dict[name]

    # Execute operations sequentially
    for op in compiled.ops:
        # Get left operand
        if op.left_idx >= 0:
            left = workspace[op.left_idx]
        else:
            left = op.left_const

        # Get right operand
        if op.right_idx >= 0:
            right = workspace[op.right_idx]
        elif op.right_const is not None:
            right = op.right_const
        else:
            right = None

        # Execute operation
        if op.op in _BINARY_OPS:
            result = _BINARY_OPS[op.op](left, right)
        elif op.op in _UNARY_OPS:
            result = _UNARY_OPS[op.op](left)
        elif op.op == "getitem":
            result = left[op.getitem_key]
        elif op.atom_callable is not None:
            # Atom operations (norm, trace, det, etc.)
            if right is None:
                result = op.atom_callable(left)
            else:
                result = op.atom_callable(left, right)
        elif op.op == "func":
            # User-defined function
            args = [left]
            if right is not None:
                args.append(right)
            # Handle extra args for functions with >2 parameters
            if hasattr(op, '_extra_arg_indices'):
                for idx, const in zip(op._extra_arg_indices, op._extra_arg_consts):
                    if idx >= 0:
                        args.append(workspace[idx])
                    else:
                        args.append(const)
            kwargs = op.func_kwargs or {}
            result = op.func_callable(*args, **kwargs)
        elif op.op == "dict":
            # Build dict from compiled elements
            result = {}
            for key, idx, const in zip(op.dict_keys, op.elem_indices, op.elem_consts):
                if idx >= 0:
                    result[key] = workspace[idx]
                else:
                    result[key] = const
        elif op.op == "container":
            # Build list/tuple from compiled elements
            elems = []
            for idx, const in zip(op.elem_indices, op.elem_consts):
                if idx >= 0:
                    elems.append(workspace[idx])
                else:
                    elems.append(const)
            result = op.container_type(elems)
        else:
            raise NotImplementedError(f"Compiled execution for op: {op.op}")

        workspace[op.result_idx] = result

    return workspace[compiled.result_idx]


# =============================================================================
# CODEGEN COMPILER - Generates actual Python code for maximum speed
# =============================================================================

class CodegenCompiler:
    """
    Compiles expression trees into Python source code.

    This is the fastest approach as it generates actual Python code that
    can be compiled once and executed at native Python speed, with no
    interpreter overhead per operation.
    """

    def __init__(self):
        self.code_lines: List[str] = []
        self.temp_counter = 0
        self.expr_to_temp: Dict[int, str] = {}
        self.var_names: List[str] = []
        self.var_name_to_idx: Dict[str, int] = {}  # O(1) lookup for variable indices
        self.constants: Dict[str, Any] = {}
        self.const_counter = 0
        self.atoms: Dict[str, Callable] = {}
        self.atom_counter = 0
        self.funcs: Dict[str, Tuple[Callable, Dict]] = {}
        self.func_counter = 0

    def _alloc_temp(self) -> str:
        """Allocate a new temporary variable name."""
        name = f"_t{self.temp_counter}"
        self.temp_counter += 1
        return name

    def _alloc_const(self, value: Any) -> str:
        """Allocate a constant and return its reference name."""
        name = f"_c{self.const_counter}"
        self.const_counter += 1
        self.constants[name] = value
        return name

    def _alloc_atom(self, atom: Callable) -> str:
        """Allocate an atom callable and return its reference name."""
        name = f"_atom{self.atom_counter}"
        self.atom_counter += 1
        self.atoms[name] = atom
        return name

    def _alloc_func(self, func: Callable, kwargs: Dict) -> str:
        """Allocate a function callable and return its reference name."""
        name = f"_func{self.func_counter}"
        self.func_counter += 1
        self.funcs[name] = (func, kwargs)
        return name

    def compile(self, expr) -> Callable:
        """
        Compile an expression tree into a callable Python function.

        Returns:
            A function that takes (var_dict) and returns the result.
        """
        self.code_lines = []
        self.temp_counter = 0
        self.expr_to_temp = {}
        self.var_names = []
        self.var_name_to_idx = {}
        self.constants = {}
        self.const_counter = 0
        self.atoms = {}
        self.atom_counter = 0
        self.funcs = {}
        self.func_counter = 0

        result_name = self._emit_node(expr)

        # Build list of constants/atoms/funcs for direct indexing (faster than dict)
        const_list = [self.constants[f"_c{i}"] for i in range(self.const_counter)]
        atom_list = [self.atoms[f"_atom{i}"] for i in range(self.atom_counter)]
        func_list = [self.funcs[f"_func{i}"] for i in range(self.func_counter)]

        # Build variable extraction code at start of function
        var_extractions = []
        for i, name in enumerate(self.var_names):
            var_extractions.append(f"_v{i} = _vars['{name}']")

        # Build the function with list indexing instead of dict lookups
        extraction_code = "\n    ".join(var_extractions) if var_extractions else "pass"
        func_body = "\n    ".join(self.code_lines)
        func_code = f"""def _compiled_eval(_vars, _C, _A, _F):
    {extraction_code}
    {func_body}
    return {result_name}
"""

        # Compile and execute to get the function
        local_ns = {}
        exec(func_code, {"__builtins__": {}}, local_ns)
        inner_func = local_ns["_compiled_eval"]

        # Capture lists in closure (tuple is slightly faster)
        const_tuple = tuple(const_list)
        atom_tuple = tuple(atom_list)
        func_tuple = tuple(func_list)

        def compiled_eval(var_dict):
            return inner_func(var_dict, const_tuple, atom_tuple, func_tuple)

        # Store metadata for debugging
        compiled_eval._source = func_code
        compiled_eval._var_names = list(self.var_names)

        return compiled_eval

    def _emit_node(self, expr) -> str:
        """
        Emit code for a node and return the variable name holding the result.
        """
        # Check if already emitted (CSE)
        expr_id = id(expr)
        if expr_id in self.expr_to_temp:
            return self.expr_to_temp[expr_id]

        if isinstance(expr, Variable):
            # Variables come from var_dict, extracted at function start
            if expr.name not in self.var_name_to_idx:
                var_idx = len(self.var_names)
                self.var_names.append(expr.name)
                self.var_name_to_idx[expr.name] = var_idx
            else:
                var_idx = self.var_name_to_idx[expr.name]
            result = f"_v{var_idx}"
            self.expr_to_temp[expr_id] = result
            return result

        elif isinstance(expr, Function):
            # Emit all arguments
            arg_names = [self._emit_node(arg) for arg in expr.args]

            # Allocate function reference
            kwargs = expr.kwargs if hasattr(expr, 'kwargs') else {}
            func_idx = self.func_counter
            self._alloc_func(expr.func, kwargs)

            temp = self._alloc_temp()
            args_str = ", ".join(arg_names)

            if kwargs:
                self.code_lines.append(f"{temp} = _F[{func_idx}][0]({args_str}, **_F[{func_idx}][1])")
            else:
                self.code_lines.append(f"{temp} = _F[{func_idx}][0]({args_str})")

            self.expr_to_temp[expr_id] = temp
            return temp

        elif isinstance(expr, Expr):
            # Emit left operand
            left_name = self._emit_node(expr.left)

            # Emit right operand if present
            if expr.right is not None and expr.op != "getitem":
                right_name = self._emit_node(expr.right)
            else:
                right_name = None

            temp = self._alloc_temp()

            # Generate code based on operation
            if expr.op == "add":
                self.code_lines.append(f"{temp} = {left_name} + {right_name}")
            elif expr.op == "sub":
                self.code_lines.append(f"{temp} = {left_name} - {right_name}")
            elif expr.op == "mul":
                self.code_lines.append(f"{temp} = {left_name} * {right_name}")
            elif expr.op == "div":
                self.code_lines.append(f"{temp} = {left_name} / {right_name}")
            elif expr.op == "pow":
                self.code_lines.append(f"{temp} = {left_name} ** {right_name}")
            elif expr.op == "matmul":
                self.code_lines.append(f"{temp} = {left_name} @ {right_name}")
            elif expr.op == "neg":
                self.code_lines.append(f"{temp} = -{left_name}")
            elif expr.op == "transpose":
                self.code_lines.append(f"{temp} = {left_name}.T")
            elif expr.op == "flatten":
                self.code_lines.append(f"{temp} = {left_name}.flatten()")
            elif expr.op == "getitem":
                # Store the key as a constant (use index for tuple access)
                key_idx = self.const_counter
                self._alloc_const(expr.right)
                self.code_lines.append(f"{temp} = {left_name}[_C[{key_idx}]]")
            elif callable(expr):
                # Atom operation (norm, trace, etc.)
                atom_idx = self.atom_counter
                self._alloc_atom(expr)
                if right_name is None:
                    self.code_lines.append(f"{temp} = _A[{atom_idx}]({left_name})")
                else:
                    self.code_lines.append(f"{temp} = _A[{atom_idx}]({left_name}, {right_name})")
            else:
                raise NotImplementedError(f"Codegen for op: {expr.op}")

            self.expr_to_temp[expr_id] = temp
            return temp

        elif isinstance(expr, dict):
            # Compile dict - emit each value
            result_parts = []
            for k, v in expr.items():
                v_name = self._emit_node(v)
                key_idx = self.const_counter
                self._alloc_const(k)
                result_parts.append(f"_C[{key_idx}]: {v_name}")

            temp = self._alloc_temp()
            self.code_lines.append(f"{temp} = {{{', '.join(result_parts)}}}")
            self.expr_to_temp[expr_id] = temp
            return temp

        elif isinstance(expr, (list, tuple)) and not isinstance(expr, np.ndarray):
            # Compile list/tuple
            elem_names = [self._emit_node(e) for e in expr]
            temp = self._alloc_temp()

            if isinstance(expr, tuple):
                self.code_lines.append(f"{temp} = ({', '.join(elem_names)},)")
            else:
                self.code_lines.append(f"{temp} = [{', '.join(elem_names)}]")

            self.expr_to_temp[expr_id] = temp
            return temp

        else:
            # Constant - store in constants tuple (use index for access)
            const_idx = self.const_counter
            self._alloc_const(expr)
            return f"_C[{const_idx}]"


def compile_to_function(expr) -> Callable:
    """
    Compile an expression tree into a Python function.

    This is the most optimized compilation method - it generates
    actual Python code that runs at native speed.

    Args:
        expr: An expression tree (Expr, Variable, Function, or constant)

    Returns:
        A callable that takes a var_dict and returns the result

    Example:
        >>> x = Variable((3,), name="x")
        >>> y = Variable((3,), name="y")
        >>> expr = x + y * 2
        >>> f = compile_to_function(expr)
        >>> result = f({"x": np.array([1,2,3]), "y": np.array([4,5,6])})
    """
    compiler = CodegenCompiler()
    return compiler.compile(expr)


# Cache for codegen-compiled functions
_codegen_cache: Dict[int, Callable] = {}


def eval_expression_codegen(expr, var_dict, use_value=False):
    """
    Evaluate an expression using codegen compilation.

    This is the fastest evaluation method, caching compiled functions.

    Args:
        expr: Expression tree to evaluate
        var_dict: Dictionary mapping variable names to values
        use_value: If True, use Variable._value (falls back to original impl)

    Returns:
        Evaluated result as numpy array
    """
    if use_value:
        from .parser import eval_expression
        return eval_expression(expr, var_dict, use_value=True)

    expr_id = id(expr)
    if expr_id not in _codegen_cache:
        _codegen_cache[expr_id] = compile_to_function(expr)

    return _codegen_cache[expr_id](var_dict)


def clear_codegen_cache():
    """Clear the codegen compilation cache."""
    global _codegen_cache
    _codegen_cache = {}


# =============================================================================
# Unified Compilation Interface
# =============================================================================

def compile_expr(expr, strategy: str = "codegen") -> Callable:
    """
    Compile an expression tree using the specified strategy.
    
    This is the recommended unified interface for expression compilation.
    It allows you to choose between different compilation strategies based
    on your performance and debugging needs.
    
    Args:
        expr: An expression tree (Expr, Variable, Function, or constant)
        strategy: Compilation strategy to use:
            - "codegen" (default): Generates Python source code for maximum speed.
              Best for expressions evaluated many times in optimization loops.
            - "interpreted": Compiles to an operation sequence executed by an
              interpreter loop. Easier to debug and inspect.
    
    Returns:
        A callable that takes a var_dict and returns the result.
        The callable signature is: compiled(var_dict: Dict[str, np.ndarray]) -> np.ndarray
    
    Example:
        >>> x = Variable((3,), name="x")
        >>> y = Variable((3,), name="y")
        >>> expr = x + y * 2
        >>> 
        >>> # Use codegen (fastest, default)
        >>> f = compile_expr(expr)
        >>> result = f({"x": np.array([1,2,3]), "y": np.array([4,5,6])})
        >>>
        >>> # Use interpreted (for debugging)
        >>> f = compile_expr(expr, strategy="interpreted")
        >>> result = f({"x": np.array([1,2,3]), "y": np.array([4,5,6])})
    
    Performance Notes:
        - For single evaluations, the overhead of compilation may exceed
          the benefits. Consider using the parser's eval_expression() directly.
        - For repeated evaluations (100+), codegen provides significant speedup.
        - The interpreted strategy is ~2-3x faster than recursive evaluation
          but ~2-3x slower than codegen.
    
    Raises:
        ValueError: If an invalid strategy is specified.
    """
    if strategy == "codegen":
        return compile_to_function(expr)
    elif strategy == "interpreted":
        return compile_expression(expr)
    else:
        raise ValueError(
            f"Invalid compilation strategy '{strategy}'. "
            f"Valid options: 'codegen', 'interpreted'"
        )



import autograd.numpy as np

from .constraint import Constraint
from .set import Set
from .constants import Curvature as C


# Type alias for values that can be used with expressions
ExprLike = "BaseExpr | np.ndarray | int | float | complex"


def broadcast_shapes(left_shape: tuple, right_shape: tuple) -> tuple:
    """Compute the result shape of broadcasting two shapes together (NumPy rules)."""
    if not left_shape or not right_shape:
        return left_shape if not right_shape else right_shape

    left_shape = list(left_shape)
    right_shape = list(right_shape)

    while len(left_shape) < len(right_shape):
        left_shape.insert(0, 1)
    while len(right_shape) < len(left_shape):
        right_shape.insert(0, 1)

    result_shape = []
    for left_dim, right_dim in zip(left_shape, right_shape):
        if left_dim == 1 or right_dim == 1:
            result_shape.append(max(left_dim, right_dim))
        elif left_dim == right_dim:
            result_shape.append(left_dim)
        else:
            raise ValueError(f"Incompatible shapes for broadcasting: {tuple(left_shape)}, {tuple(right_shape)}")

    return tuple(result_shape)


def expr_to_str(expr: ExprLike) -> str:
    """Convert an expression-like object to a string representation."""
    if isinstance(expr, BaseExpr):
        return str(expr)
    elif isinstance(expr, np.ndarray):
        return f"Const(shape={expr.shape}, id={id(expr)})"
    elif np.isscalar(expr):
        return f"Const(type={type(expr).__name__}, id={id(expr)})"
    return f"Unknown({type(expr).__name__})"


class BaseExpr:
    """Base class for all expression types in nvxpy.

    Provides arithmetic operators (+, -, *, /, @, **) that build expression trees,
    comparison operators (>=, <=, ==) that create constraints, and matrix operators
    (>>, <<) for semidefinite constraints.

    The ^ operator creates discrete set membership constraints when used with a
    DiscreteSet or list of values.
    """

    __array_priority__ = 100

    @property
    def T(self):
        return Expr("transpose", self)
    
    def flatten(self):
        return Expr("flatten", self)

    def __add__(self, other):
        return Expr("add", self, other)

    def __radd__(self, other):
        return Expr("add", other, self)

    def __sub__(self, other):
        return Expr("sub", self, other)

    def __rsub__(self, other):
        return Expr("sub", other, self)

    def __mul__(self, other):
        return Expr("mul", self, other)

    def __rmul__(self, other):
        return Expr("mul", other, self)

    def __matmul__(self, other):
        return Expr("matmul", self, other)

    def __rmatmul__(self, other):
        return Expr("matmul", other, self)

    def __truediv__(self, other):
        return Expr("div", self, other)
    
    def __rtruediv__(self, other):
        return Expr("div", other, self)

    def __pow__(self, other):
        return Expr("pow", self, other)

    def __neg__(self):
        return Expr("neg", self)

    def __getitem__(self, key):
        return Expr("getitem", self, key)

    def __ge__(self, other):
        return Constraint(self, ">=", other)

    def __le__(self, other):
        return Constraint(self, "<=", other)

    def __eq__(self, other):
        return Constraint(self, "==", other)

    def __rshift__(self, other):
        return Constraint(self, ">>", other)

    def __lshift__(self, other):
        return Constraint(self, "<<", other)

    def __xor__(self, other):
        if isinstance(other, (list, tuple)):
            from .sets.discrete_set import _coerce_to_discrete_set
            other = _coerce_to_discrete_set(other)
        if not isinstance(other, Set):
            raise TypeError("Right operand of ^ must be a Set object")
        return other.constrain(self)


class Expr(BaseExpr):
    """An expression node in the computation graph.

    Represents an operation applied to one or two operands. Operations include
    arithmetic (add, sub, mul, div, pow, neg), matrix operations (matmul, transpose),
    and indexing (getitem).

    Attributes:
        op: The operation name (e.g., "add", "mul", "matmul")
        left: The left operand (or only operand for unary operations)
        right: The right operand (None for unary operations)
    """

    def __init__(self, op: str, left, right=None):
        self.op = op
        self.left = left
        self.right = right

    def __repr__(self):
        left_str = expr_to_str(self.left)
        right_str = None
        if self.right is not None:
            right_str = expr_to_str(self.right)
        return f"{self.op}({left_str}, {right_str})" if right_str else f"{self.op}({left_str})"
    
    def __hash__(self):
        return hash(str(self))

    @property
    def value(self):
        from .parser import eval_expression
        return eval_expression(self, None, use_value=True)
    
    @property
    def curvature(self):
        from .variable import Variable

        def negate_curvature(conv):
            if conv == C.CONVEX:
                return C.CONCAVE
            elif conv == C.CONCAVE:
                return C.CONVEX
            else:
                return conv
            
        if isinstance(self.left, Expr) or isinstance(self.left, Variable):
            left = self.left.curvature
            if left == C.CONSTANT:
                left_val = self.left.value
        else:
            left = C.CONSTANT
            left_val = self.left
        if isinstance(self.right, Expr) or isinstance(self.right, Variable):
            right = self.right.curvature
            if right == C.CONSTANT:
                right_val = self.right.value
        elif self.right is not None:
            right = C.CONSTANT
            right_val = self.right
        else:
            right = None
            right_val = None

        if self.op in ("add", "sub"):
            if self.op == "sub":
                right = negate_curvature(right)
            if left == right:
                return left
            elif C.CONSTANT in (left, right):
                return left if right == C.CONSTANT else right
            elif C.AFFINE in (left, right):
                return left if right == C.AFFINE else right
            else:
                return C.UNKNOWN

        elif self.op == "mul":
            if left == C.CONSTANT:
                if all([left_val >= 0]):
                    return right
                elif all([left_val < 0]):
                    return negate_curvature(right)
            elif right == C.CONSTANT:
                if all([right_val >= 0]):
                    return left
                elif all([right_val < 0]):
                    return negate_curvature(left)
            return C.UNKNOWN

        elif self.op == "div":
            if right == C.CONSTANT and all([right_val >= 0]):
                return left
            elif right == C.CONSTANT and all([right_val < 0]):
                return negate_curvature(left)
            else:
                return C.UNKNOWN

        elif self.op == "pow":
            if right == C.CONSTANT:
                if right_val == 1:
                    return left
                elif right_val > 1 and left in (C.CONVEX, C.AFFINE):
                    return C.CONVEX
                elif 0 < right_val < 1 and left == C.CONCAVE:
                    return C.CONCAVE
                else:
                    return C.UNKNOWN
            else:
                return C.UNKNOWN

        elif self.op == "neg":
            return negate_curvature(left)

        elif self.op == "matmul":
            if left == C.AFFINE and right == C.CONSTANT:
                return C.AFFINE
            elif left == C.CONSTANT and right == C.AFFINE:
                return C.AFFINE
            else:
                return C.UNKNOWN

        elif self.op == "getitem":
            return left

        elif self.op == "transpose":
            return left
        
        elif self.op == "flatten":
            return left
        
        raise NotImplementedError(f"Curvature of Expr: {self.op} is not implemented")

    @property
    def shape(self):
        from .variable import Variable

        def get_shape(obj):
            if isinstance(obj, (Expr, Variable)):
                return obj.shape
            elif isinstance(obj, np.ndarray):
                return obj.shape
            elif np.isscalar(obj):
                return ()
            elif isinstance(obj, (slice, tuple)):
                return None
            else:
                raise ValueError(f"Cannot determine shape of {type(obj)}")

        left_shape = get_shape(self.left)
        right_shape = get_shape(self.right) if self.right is not None else None

        if self.op in ("add", "sub", "mul", "div"):
            return broadcast_shapes(left_shape, right_shape)

        elif self.op == "pow":
            return left_shape

        elif self.op == "neg":
            return left_shape

        elif self.op == "matmul":
            if len(left_shape) == 1 and len(right_shape) == 1:
                if left_shape[0] != right_shape[0]:
                    raise ValueError(f"Incompatible shapes for matmul: {left_shape}, {right_shape}")
                return ()
            elif len(left_shape) == 1:
                left_shape = (1,) + left_shape
            elif len(right_shape) == 1:
                right_shape = right_shape + (1,)
            
            if left_shape[-1] != right_shape[-2]:
                raise ValueError(f"Incompatible shapes for matmul: {left_shape}, {right_shape}")
            
            left_batch = left_shape[:-2]
            right_batch = right_shape[:-2]
            
            max_batch_dims = max(len(left_batch), len(right_batch))
            left_batch = (1,) * (max_batch_dims - len(left_batch)) + left_batch
            right_batch = (1,) * (max_batch_dims - len(right_batch)) + right_batch
            
            batch_shape = []
            for left_dim, right_dim in zip(left_batch, right_batch):
                if left_dim == 1 or right_dim == 1:
                    batch_shape.append(max(left_dim, right_dim))
                elif left_dim == right_dim:
                    batch_shape.append(left_dim)
                else:
                    raise ValueError(f"Incompatible batch dimensions for matmul: {left_shape}, {right_shape}")
            
            result_shape = tuple(batch_shape) + (left_shape[-2], right_shape[-1])
            
            if len(left_shape) == 1:
                result_shape = result_shape[1:]
            
            return result_shape

        elif self.op == "getitem":
            if isinstance(self.right, tuple):
                new_shape = []
                orig_shape = list(left_shape)
                dim_idx = 0
                for idx in self.right:
                    if dim_idx >= len(orig_shape):
                        raise IndexError(f"too many indices for array of dimension {len(orig_shape)}")
                    dim_size = orig_shape[dim_idx]
                    if isinstance(idx, slice):
                        start = 0 if idx.start is None else idx.start
                        stop = dim_size if idx.stop is None else idx.stop
                        step = 1 if idx.step is None else idx.step
                        if step == 0:
                            raise ValueError("Slice step cannot be zero")
                        # Normalize negative indices
                        if start < 0:
                            start = max(0, dim_size + start)
                        if stop < 0:
                            stop = max(0, dim_size + stop)
                        # Clamp to valid range
                        start = min(start, dim_size)
                        stop = min(stop, dim_size)
                        new_shape.append(max(0, (stop - start + step - 1) // step) if step > 0 else max(0, (start - stop - step - 1) // (-step)))
                        dim_idx += 1
                    elif isinstance(idx, int):
                        # Validate index bounds
                        if idx < -dim_size or idx >= dim_size:
                            raise IndexError(f"index {idx} is out of bounds for axis {dim_idx} with size {dim_size}")
                        dim_idx += 1
                        continue
                    else:
                        raise ValueError(f"Unsupported index type: {type(idx)}")
                return tuple(new_shape) if new_shape else (1,)
            elif isinstance(self.right, (int, slice)):
                if isinstance(self.right, int):
                    dim_size = left_shape[0] if left_shape else 0
                    if self.right < -dim_size or self.right >= dim_size:
                        raise IndexError(f"index {self.right} is out of bounds for axis 0 with size {dim_size}")
                    return left_shape[1:] if len(left_shape) > 1 else (1,)
                else:
                    dim_size = left_shape[0] if left_shape else 0
                    start = 0 if self.right.start is None else self.right.start
                    stop = dim_size if self.right.stop is None else self.right.stop
                    step = 1 if self.right.step is None else self.right.step
                    if step == 0:
                        raise ValueError("Slice step cannot be zero")
                    # Normalize negative indices
                    if start < 0:
                        start = max(0, dim_size + start)
                    if stop < 0:
                        stop = max(0, dim_size + stop)
                    # Clamp to valid range
                    start = min(start, dim_size)
                    stop = min(stop, dim_size)
                    slice_len = max(0, (stop - start + step - 1) // step) if step > 0 else max(0, (start - stop - step - 1) // (-step))
                    return (slice_len,) + left_shape[1:]
            else:
                raise ValueError(f"Unsupported index type: {type(self.right)}")

        elif self.op == "transpose":
            return left_shape[::-1]
        
        elif self.op == "flatten":
            return (np.prod(left_shape),)
        
        raise NotImplementedError(f"Shape inference for operation {self.op} is not implemented")



"""
Branch-and-Bound MINLP Solver

This package implements a comprehensive branch-and-bound algorithm for
mixed-integer nonlinear programming (MINLP).

Modules:
- backend: Main BranchAndBoundBackend solver class
- node: Node, statistics, and variable info dataclasses
- branching: Variable branching strategies
- heuristics: Primal heuristics for finding feasible solutions
- cuts: Outer approximation cut generation and management
- utils: Shared utility functions
"""

from .backend import BranchAndBoundBackend
from .node import (
    BBNode,
    BBStats,
    BranchingStrategy,
    DiscreteVarInfo,
    NodeSelection,
    PseudocostData,
)
from .cuts import OACut, generate_oa_cuts, prune_cut_pool

__all__ = [
    "BranchAndBoundBackend",
    "BBNode",
    "BBStats",
    "BranchingStrategy",
    "DiscreteVarInfo",
    "NodeSelection",
    "PseudocostData",
    "OACut",
    "generate_oa_cuts",
    "prune_cut_pool",
]




from __future__ import annotations

import autograd.numpy as np

from ..expression import Expr, BaseExpr, ExprLike
from ..constants import Curvature as C, EPSILON


def _smooth_norm(
    x: np.ndarray, ord: int | float | str = 2, axis: int | None = None
) -> np.ndarray:
    """
    Compute norm with epsilon smoothing to avoid gradient singularity at zero.

    The gradient of ||x|| is x / ||x||, which is undefined when ||x|| = 0.
    We compute sqrt(||x||^2 + eps) instead, which has a well-defined gradient
    everywhere: x / sqrt(||x||^2 + eps).
    """
    if ord == 2 and axis is None:
        # L2 norm (default): sqrt(sum(x^2) + eps)
        return np.sqrt(np.sum(x * x) + EPSILON)
    elif ord == "fro":
        # Frobenius norm: sqrt(sum(x^2) + eps)
        return np.sqrt(np.sum(x * x) + EPSILON)
    elif ord == 1 and axis is None:
        # L1 norm: sum(|x|) - no singularity issue, but use smooth abs
        return np.sum(np.sqrt(x * x + EPSILON))
    elif ord == np.inf and axis is None:
        # Inf norm: max(|x|) - use logsumexp approximation for smoothness
        return np.max(np.sqrt(x * x + EPSILON))
    else:
        # Fall back to numpy for other norms, with post-hoc smoothing
        raw_norm = np.linalg.norm(x, ord=ord, axis=axis)
        return np.sqrt(raw_norm * raw_norm + EPSILON)


class norm(Expr):
    """Vector or matrix norm.

    norm(x, ord=2) computes the L2 norm by default.
    Supports ord=1, 2, inf, 'fro', 'nuc'.

    Curvature rules:
    - For ord >= 1: convex for affine/convex arguments
    - For ord < 1: generally non-convex
    """

    def __init__(
        self, left: ExprLike, ord: int | float | str = 2, axis: int | None = None
    ) -> None:
        self.ord = ord
        self.axis = axis
        super().__init__("norm", left)

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return _smooth_norm(x, ord=self.ord, axis=self.axis)

    @property
    def curvature(self) -> C:
        arg = self.left
        arg_cvx = (
            arg.curvature
            if isinstance(arg, BaseExpr)
            else C.CONSTANT
        )

        if arg_cvx == C.CONSTANT:
            arg_cvx = C.AFFINE

        o = self.ord
        is_convex_ord = False

        if o is None:
            is_convex_ord = True

        elif isinstance(o, (int, float)):
            if o >= 1 or np.isposinf(o):
                is_convex_ord = True

        elif isinstance(o, str):
            if o in ("fro", "nuc"):
                is_convex_ord = True

        if is_convex_ord and arg_cvx in (C.AFFINE, C.CONVEX):
            return C.CONVEX

        return C.UNKNOWN

    @property
    def shape(self) -> tuple[int, ...]:
        if self.axis is None:
            return (1,)
        arg_shape = self.left.shape if isinstance(self.left, BaseExpr) else np.shape(self.left)
        ndim = len(arg_shape)
        axis = self.axis if self.axis >= 0 else ndim + self.axis
        if axis < 0 or axis >= ndim:
            raise ValueError(f"axis {self.axis} is out of bounds for array of dimension {ndim}")
        result = list(arg_shape)
        result.pop(axis)
        return tuple(result) if result else (1,)


from __future__ import annotations

import autograd.numpy as np

from ..expression import Expr, ExprLike
from ..constants import Curvature as C


class det(Expr):
    """Matrix determinant.

    det(A) returns the determinant of square matrix A.
    Generally non-convex.
    """

    def __init__(self, left: ExprLike) -> None:
        super().__init__("det", left)

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return np.linalg.det(x)

    @property
    def curvature(self) -> C:
        return C.UNKNOWN

    @property
    def shape(self) -> tuple[int, ...]:
        return (1,)


from __future__ import annotations

import autograd.numpy as np

from ..expression import Expr, BaseExpr, ExprLike
from ..constants import Curvature as C


class trace(Expr):
    """Sum along the diagonal of a matrix.

    trace(X) returns the sum of diagonal elements.
    Preserves curvature (affine operation).
    """

    def __init__(self, left: ExprLike, offset: int = 0) -> None:
        self.offset = offset
        super().__init__("trace", left)

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return np.trace(x, offset=self.offset)

    @property
    def curvature(self) -> C:
        arg = self.left
        arg_curv = arg.curvature if isinstance(arg, BaseExpr) else C.CONSTANT

        return arg_curv

    @property
    def shape(self) -> tuple[int, ...]:
        return (1,)


from functools import partial

import numpy as _np
import autograd.numpy as np
from autograd.extend import defvjp, primitive


_dot = partial(np.einsum, "...ij,...jk->...ik")


def _diag(a):
    return np.eye(a.shape[-1]) * a


def T(x):
    return np.swapaxes(x, -1, -2)


@primitive
def svd(a, full_matrices=True, compute_uv=True):
    return _np.linalg.svd(a, full_matrices=full_matrices, compute_uv=compute_uv)


# From: https://github.com/HIPS/autograd/blob/master/autograd/numpy/linalg.py
# Manually defined here to make gradient stable for diagonal matrices
def grad_svd(usv_, a, full_matrices=True, compute_uv=True):
    def vjp(g):
        usv = usv_

        if not compute_uv:
            s = usv

            # Need U and V so do the whole svd anyway...
            usv = svd(a, full_matrices=False)
            u = usv[0]
            v = np.conj(T(usv[2]))

            return _dot(np.conj(u) * g[..., np.newaxis, :], T(v))

        elif full_matrices:
            raise NotImplementedError(
                "Gradient of svd not implemented for full_matrices=True"
            )

        else:
            u = usv[0]
            s = usv[1]
            v = np.conj(T(usv[2]))

            m, n = a.shape[-2:]

            k = np.min((m, n))
            # broadcastable identity array with shape (1, 1, ..., 1, k, k)
            i = np.reshape(
                np.eye(k), np.concatenate((np.ones(a.ndim - 2, dtype=int), (k, k)))
            )

            f = 1 / (
                (s[..., np.newaxis, :] ** 2 - s[..., :, np.newaxis] ** 2 + i) + 1e-8
            )  # <---- change from original

            gu = g[0]
            gs = g[1]
            gv = np.conj(T(g[2]))

            utgu = _dot(T(u), gu)
            vtgv = _dot(T(v), gv)
            t1 = (f * (utgu - np.conj(T(utgu)))) * s[..., np.newaxis, :]
            t1 = t1 + i * gs[..., :, np.newaxis]
            t1 = t1 + s[..., :, np.newaxis] * (f * (vtgv - np.conj(T(vtgv))))

            if np.iscomplexobj(u):
                t1 = t1 + 1j * np.imag(_diag(utgu)) / s[..., np.newaxis, :]

            t1 = _dot(_dot(np.conj(u), t1), T(v))

            if m < n:
                i_minus_vvt = np.reshape(
                    np.eye(n), np.concatenate((np.ones(a.ndim - 2, dtype=int), (n, n)))
                ) - _dot(v, np.conj(T(v)))
                t1 = t1 + np.conj(
                    _dot(_dot(u / s[..., np.newaxis, :], T(gv)), i_minus_vvt)
                )

                return t1

            elif m == n:
                return t1

            elif m > n:
                i_minus_uut = np.reshape(
                    np.eye(m), np.concatenate((np.ones(a.ndim - 2, dtype=int), (m, m)))
                ) - _dot(u, np.conj(T(u)))
                t1 = t1 + T(_dot(_dot(v / s[..., np.newaxis, :], T(gu)), i_minus_uut))

                return t1

    return vjp


defvjp(svd, grad_svd)



from __future__ import annotations

import autograd.numpy as np

from ..expression import Expr, BaseExpr, ExprLike
from ..constants import Curvature as C


class exp(Expr):
    """Exponential function.

    exp(x) is convex.

    Curvature rules:
    - exp(constant) -> constant
    - exp(affine) -> convex
    - exp(convex) -> convex (composition rule: convex + nondecreasing)
    - exp(concave) -> unknown
    """

    def __init__(self, left: ExprLike) -> None:
        super().__init__("exp", left)

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return np.exp(x)

    @property
    def curvature(self) -> C:
        arg = self.left
        arg_cvx = arg.curvature if isinstance(arg, BaseExpr) else C.CONSTANT

        if arg_cvx == C.CONSTANT:
            return C.CONSTANT

        if arg_cvx in (C.AFFINE, C.CONVEX):
            # exp is convex and nondecreasing, so exp(convex) is convex
            return C.CONVEX

        return C.UNKNOWN

    @property
    def shape(self) -> tuple[int, ...]:
        if isinstance(self.left, BaseExpr):
            return self.left.shape
        return np.shape(self.left) or (1,)
